
! Omcatalogue version of omsrclistcomb
!+
! Module: omsrclistcomb_mod
!
! Authors:  Christopher Brindle - MSSL
!           Andrew King - MSSL
!           Tim Harries - MSSL
!          
! Description: Combines OSW source list into an observation source list
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History: 
   
!
!************************************************************************************************
module omsrclistcomb_mod
  use param
  use cal
  use dal
  use Errorhandling
  use sla
  use omallocation_mod
  use omconstants_mod
  use omastromcorrect_mod
  use omutility_mod
  use omgraph_mod
!  use fluxfit_mod
  use omsourceflags_mod
  use OmSourceParam_mod    
  use sourcetable_mod
  
  implicit none 

  private 

  public omsrclistcomb

  !*****************************
  ! Structure to define a color
  !*****************************
  type ColourT
     integer :: mag1
     integer :: mag2
     character(20) label, label1
     character(20) elabel
     character(20) flabel
  end type ColourT

!******************************************************************
! The OM filters
!******************************************************************
integer(int32), private :: filt(7) =  (/ UVW2, UVM2, UVW1, U, B, V, WHITE /)

integer(int8), private,  pointer     :: qflagTable(:, :) => null()        ! Quality flag for each filter
integer(int8), private,  pointer     :: cflagTable(:, :) => null()        ! Confusion flag table
integer(int8), private,  pointer     :: eflagTable(:, :) => null()        ! Extension flag table

real(single), private, pointer       :: mag_table(:, :) => null()         ! table of magnitudes
real(single), private, pointer       :: magerr_table(:, :) => null()      ! table of magnitude errors
real(single), private, pointer       :: corrRate_table(:, :) => null()    ! table of magnitudes
real(single), private, pointer       :: corrRateErr_table(:, :) => null() ! table of magnitude error
real(single), private, pointer       :: majorTable(:,:) => null()         ! table of major axes
real(single), private, pointer       :: minorTable(:,:) => null()         ! table of major axes
real(single), private, pointer       :: paTable(:,:) => null()            ! table of major axes
real(single), private, pointer       :: significanceTable(:, :) => null() ! table of magnitudes
real(double), private :: raz, decz
real(single), private :: nsigma
logical, private, pointer        :: magexist(:,:) => null()               ! does a magnitude exist?


!type(SourceData), pointer :: omFitData(:) => null() ! Stores the data needed for the template-spectrum fitting
 
real(single), parameter :: tolerance = 2.0 ! was 2
integer(int32), parameter, private               :: int32NullValue = -1      

contains 

!****************************************
! The main program
!****************************************
subroutine omsrclistcomb

    integer, parameter          :: nFilt = 7             ! no of filters
    integer, parameter          :: nCol = 15             ! no of colours for instrumental magnitudes
    integer, parameter          :: nCol1 = 8             ! no of colours for standard magnitudes
    type(ColourT)               :: instrumentalColour(nCol)
    type(ColourT)               :: standardColour(nCol)

    type(DataSetT)              :: oswDataSet            ! handles
    type(DataSetT)              :: outDataSet            ! handle for output dataset
    type(TableT)                :: tableSet
  
    character(10), pointer      :: filter(:) => null()             ! filters for osw source lists
    character(10), pointer      :: sourcefilter(:) => null()       ! filters for osw source lists

    character(1024), pointer    :: oswFile(:) => null()             ! filenames
    character(50)               :: label                  ! column label
    integer                     :: nStart                 ! start no of sources
    integer                     :: nIndep                 ! number of independent sources
    integer                     :: nosw                   ! no of osw lists
    integer                     :: i, j, k, n             ! counters
    integer                     :: nSources               ! total no of sources
    integer(int32), pointer     :: nrows(:)               ! no of rows in each osw source list
    integer(int32), pointer     :: oswid(:) => null()                ! no of each osw
 
    real(single), pointer       :: magColour(:, :) => null()   
    real(single), pointer       :: magColourError(:, :) => null()   
    real(single), pointer       :: standardMagnitude(:, :) => null()  
    real(single), pointer       :: standardMagnitudeError(:, :) => null()
    real(single), pointer       :: stdMagnitudeStore(:, :, :) => null()  
    real(single), pointer       :: stdMagnitudeErrorStore(:, :, :) => null()
   
    real(single), pointer       :: magnitude(:, :) => null()   
    real(single), pointer       :: magnitudeError(:, :) => null()   

    real(double), pointer       :: galong(:) => null()      ! array for galactic longitudes
    real(double), pointer       :: galat(:) => null()       ! array for galactic latitudes    
    real(double)                :: stdColour
    real(double)                :: stdMagnitude1
    real(double)                :: stdMagnitude2
    real(double)                :: stdColourError
    real(double)                :: stdMagnitude1Error
    real(double)                :: stdMagnitude2Error
      
    integer                     :: noColours
    character(100)              :: colourString(15), stdString(15)
    character(100)              :: colourErrorString(15)
    character(200)              :: filterString, stdString1, stdColourString
    real(single), pointer       :: stdColourMagnitude(: , :) => null()
    real(single), pointer       :: stdColourMagnitudeError(: , :) => null()
    integer(int32), pointer     :: RASortIndex(:) => null()
    integer                     :: iColumn, jColumn      
    integer                     :: noFilters, noFluxFilters
    integer                     :: filterIndex(20), fluxFilterIndex(20)    
    integer                     :: allocErr, position
    integer(int32), pointer     :: nCount(:, :) => null()           

    logical, dimension(1 : nCol) :: stdMagExist          ! Does a standard magnitude exist?
    real(single) :: expos    
    character(len=5)  :: filtid(nFilt) = &
               (/'UVW2 ','UVM2 ','UVW1 ','U    ','B    ','V    ', 'WHITE'/) 
   
    character(200) :: string
    type(sourceParameters), pointer :: source(:) => null()
    character(100) :: comment  
    type(sourceParameters), pointer :: newsource(:) => null() 

!type(calinfo) :: calibdata  
!integer :: testint=999
    integer, pointer :: matchedNumber(:) => null()
 

    radiansToArcSeconds = radiansToDegrees * 3600.0_double


    !************************************************
    ! Initialise the filter prority and colour 
    ! tables for the instrumental magnitudes
    !************************************************
    call SetUpInstrumentalColourTable(instrumentalColour)


    !************************************************
    ! Initialise the filter prority and colour 
    ! tables for the standard magnitudes
    !************************************************
    call SetUpStandardColourTable(standardColour)

    !************************************************
    ! Get the number of osw files and their names.
    !************************************************

    nosw = parameterCount("sourcelistsets")

    if (nosw .eq. 0) then

        call warning("Can't continue", "No source-list files to process")
        stop
    end if    

 
    oswFile => AllocateChar(1024, nosw, "oswFile")

    nsigma = realParameter("nsigma")

    nrows => AllocateInt32(1, nosw, "nrows")   ! The number of rows in each source-list file
 
    filter => AllocateChar(10, nosw, "filter") ! The filter in each source-list file
    
   
    nSources = 0 
 
    !********************************************************************************************
    ! Determine total number of sources and number of source-list files with at least one source
    !********************************************************************************************
    nosw = 0
    do i = 1 , parameterCount("sourcelistsets")
                
        oswfile(i) = stringParameter("sourcelistsets", i - 1)
       
        oswDataSet = DataSet(oswFile(i),read)
      
        tableSet = table(oswDataSet, 0)
        nrows(i) = numberOfRows(tableSet)
        filter(i) = trim(stringAttribute(oswDataSet, "FILTER"))
     
        if (nrows(i) .gt. 0) then
            nSources = nSources + nrows(i)
            nosw = nosw + 1
            oswfile(nosw) = oswfile(i)
            nrows(nosw) = nrows(i)
           
        end if
        call release(tableSet)
        call release(oswDataSet)

    end do


    sourceFilter => AllocateChar(10, nsources, "filter") ! The filter of each file
    sourceFilter(1 : nrows(1)) = filter(1)
 
    k = nrows(1) + 1

    do i = 2, nosw
        sourceFilter(k : k + nrows(i) - 1) = filter(i)
        k = k + nrows(i)
    end do 

    !*************************************
    ! Check there is at least 1 source
    !*************************************
    if (nSources .eq. 0) then
        call warning("Can't continue", "No sources to process")
        stop
    end if    

    !******************************************
    ! create the output observation source list
    !******************************************
    call CreateObsSourceList(nOsw, oswFile, outDataSet, raz, decz, expos)    
    raz = raz * degreesToRadians
    decz = decz * degreesToRadians

    !******************************************
    ! Allocate memory for the arrays to hold
    ! the input data.
    !******************************************
    oswid => AllocateInt32(1, nSources, "oswid")

 
    ! The array of structs to hold data from individual source-list files
    allocate(source(1 : nSources)) 
 

    !********************************************************
    ! Read in the data from each osw source list into the 
    ! appropriate array- appending the new data to the end of
    ! that array.
    !********************************************************
     

    write(errstr,'(a)') "Loading data from source-list files"
    call message(AppMsg, Verbose)


    nStart = 0
    do i = 1 , nosw

       do j = 1, nrows(i)
        
       
           oswid(nStart + j) = i
       
       end do

       !************************************************************************
       ! Load the data from this source-list file and append to the source-list
       !***********************************************************************   
       call LoadSourceData(oswFile(i), source, nStart)
       nStart = nStart + nrows(i)  

    end do

    where (source(1 : nsources)%poserr .gt. 5.0)
        source(1 : nsources)%poserr = 5.0
    end where

    where (source%magnitude .gt. 100) 
        source%magnitude = NullValue
        source%magnitudeerr = NullValue
        source%corrcountRate = NullValue
        source%corrCountRateErr = NullValue
    end where
  
    !***************************************
    ! Do a few conversions
    !***************************************
    source%rightascension = source%rightascension * degreesToRadians
    source%declination = source%declination * degreesToRadians


    !************************************************************************************
    ! Create the merged-source-list file
    !
    !************************************************************************************
    matchedNumber => allocateint32(1, nsources, "matchednumber")

    call CreateMergedSourceList(nSources, source, nosw, oswfile,      &
                                filter, nrows, nfilt, filtId, sourceFilter, &
                                nindep, newSource, matchedNumber)


    deallocate(sourceFilter)
    nullify(sourceFilter)

    deallocate(source)
    source => newSource
   
    write(errstr,'(i6, a)') nIndep,' independent sources'
    call message(AppMsg, Verbose)

    deallocate(oswid, stat=allocErr)
    call CheckDeallocation(allocErr, "oswid")
      
    !**************************************************
    ! Create the observation source list
    ! which has nIndep rows
    !**************************************************
    tableSet = addTable(outDataSet, "SRCLIST", nIndep)

    RASortIndex => AllocateInt32(1, nSources, "RASortIndex")
 
    do i = 1, nIndep 
        RASortIndex(i) = i
    end do
    call AddInt32IndexColumn(tableSet, "SRCID", "", "Source identification", nIndep, RASortIndex, RASortIndex)
 
    !*****************************************************
    ! Find the order of increasing  RA- store in RASortIndex
    ! for adding data to source list table in  
    ! increasing RA order (ra is not sorted)
    !****************************************************
    call IndexSort(source%rightAscension, RASortIndex, nIndep)

   
    !**********************************************************
    ! Add a source ID column to each of the osw source lists
    ! telling us which source went to which entry in the
    ! observation source list 
    !**********************************************************
    call AddSourceIds(nosw, oswFile, matchedNumber, nrows, raSortIndex)

    deallocate(matchedNumber)
    nullify(matchedNumber)

 


!    call AddReal32IndexColumn(tableSet, "XPOS", "pixels", " ", nIndep, source%xpos, RASortIndex, nullValue)
!    call AddReal32IndexColumn(tableSet, "YPOS", "pixels", " ", nIndep, source%ypos, RASortIndex, nullValue)
 

    call AddReal64IndexColumn(tableSet, "RA", "degrees", "Averaged RAs of identical sources", &
                        nIndep, source%rightascension * radiansToDegrees, RASortIndex, nullValue)

    call AddReal64IndexColumn(tableSet, "DEC", "degrees", " Averaged decs of identical sources", &
                    nIndep, source%declination * radiansToDegrees, RASortIndex, nullValue)
   
!    call addradecstringcolumns(tableset, nIndep, source%rightascension * radiansToDegrees , &
!                                source%declination * radiansToDegrees, rasortindex)

    call AddReal32IndexColumn(tableSet, "POSERR", "arcsec", " ", nIndep, source%poserr, RASortIndex, nullValue)
 

 
    !**************************************
    ! Find out what filters have magnitudes
    !************************************** 
    noFilters = 0
    noFluxFilters = 0
    filterString = ""
    
    call addcomment(tableSet, "The table consists of data taken with the following filters")

  

    do i = 1 , nFilt
   
        if (ANY(magexist(1 : nIndep, i : i)))  then
            
            noFluxFilters = noFluxFilters + 1
            noFilters = noFilters + 1
            filterIndex(noFilters) = i
            write(comment, *) noFilters, ") ", trim(filtid(i))
            call addcomment(tableset, trim(comment))
        
        
        

            if (noFluxFilters .gt. 1) then
                fluxFilterIndex(noFluxFilters) = i
                label = trim(filterString)
                filterString = trim(label) // ","  // trim(filtid(i))
            else
                filterString = trim(filtid(i))
            end if
          else
        
            write(errstr,*) "No sources found in filter ", filtid(i)
            call message(AppMsg, Verbose)
        end if
    end do
 
    !******************************************************
    ! Add the significance arrays
    !****************************************************** 

 
    call AddVectorColumn(tab=tableSet, data=significanceTable, ndata=nIndep, &
                         ncols=nofilters, colindex=filterIndex, storeindex=RASortIndex, &
            heading="SIGNIF", units="", comment="Signal/noise ratio (" // trim(filterString) // ")", nullvalue=nullValue)

 
 
    deallocate(significanceTable, stat=allocErr)
    call CheckDeallocation(allocErr, "significance")



    !***********************************************************
    ! Add the mean count-rates
    !**********************************************************
    call AddVectorColumn(tab=tableSet, data=corrRate_table, ndata=nIndep, &
            ncols=noFilters, colindex=filterIndex, storeindex=RASortIndex, &
            heading="CORR_RATE", units="counts/sec", &
            comment="Mean corr-rate (" // trim(filterString) // ")", &
            nullvalue=nullValue)


    !***********************************************************
    ! Add the mean count-rate errors
    !********************************************************** 
    call AddVectorColumn(tableSet, corrRateErr_table, nIndep, noFilters, filterIndex, RASortIndex, &
            heading="CORR_RATE_ERR",  units="counts/sec",                                          &
            comment="Mean corr-rate (" // trim(filterString) // ")", &
            nullValue=nullValue)





    !*********************************************
    ! Add the instrumental magnitude and its error
    !*********************************************
    call AddVectorColumn(tableSet, mag_table, nIndep, noFilters, filterIndex, RASortIndex, &
            "INS_MAG", "mag", "Instrumental mag (" // trim(filterString) // ")", nullValue)
 
    call AddVectorColumn(tableSet, magerr_table, nIndep, noFilters, filterIndex, RASortIndex, &
            "INS_MAG_ERR", "mag", "Instrum mag error (" // trim(filterString) // ")", nullValue)
 

    !*********************************************
    ! Add the instrumental colours
    !*********************************************
    magColour => Allocate2DSingle(1, nIndep, 1, ncol, "magColour")
    magColourError => Allocate2DSingle(1, nIndep, 1, ncol, "magColourError")
    magColour = NullValue
    magColourError = NullValue

    n = 0
    string = ""

    call addcomment(tableSet, "INSTRUMENTAL COLOURS (INS_COL) - COLUMN FILTER CODES-")
   


    do i = 1, nCol     ! Loop through all the possible colours
        iColumn = instrumentalColour(i)%mag1
        jColumn = instrumentalColour(i)%mag2
        if (ANY(magexist(1 : nIndep, iColumn : iColumn)) .and.  &
            ANY(magexist(1 : nIndep, jColumn : jColumn))) then
            n = n + 1
            do k = 1, nIndep
                if (mag_table(k, iColumn) .ne. NullValue .and. &
                    mag_table(k, jColumn) .ne. NullValue) then
                    magColour(k, n) = mag_table(k, iColumn) - mag_table(k, jColumn)
                    magColourError(k, n) = AddErrors(dble(magerr_table(k, iColumn)), dble(magerr_table(k, jColumn)))
                end if
            end do 
            if (n .eq. 1) then
                string = trim(instrumentalColour(i)%label1)
            else
                label = string
                string = trim(label) // "," //   trim(instrumentalColour(i)%label1)
                
            end if

            label = trim(instrumentalColour(i)%label1) // "_INS"
            write(label, *) "Column ", n, "=", trim(instrumentalColour(i)%label1)  
            call addcomment(tableSet, trim(label))
      
 
        end if
    end do



    call AddVectorColumn(tab=tableSet, data=magColour, ndata=nIndep, ncols=n, storeindex=RASortIndex, &
            heading="INS_COL", units="mag", comment=trim(string), nullvalue=nullValue)

  


!    call AddVectorColumn(tableSet, significanceTable, nIndep, noFilters, filterIndex, RASortIndex, &
!            "SIGNIF", "sigma", " ", nullValue)


    call AddVectorColumn(tab=tableSet, data=magColourError, ndata=nIndep, ncols=n, storeindex=RASortIndex, &
            heading="INS_COL_ERR", units="mag", comment=trim(string), nullvalue=nullValue)
 

    !*******************************************************************
    ! now we have the colours
    ! we can currently have (U-B) and/or (B-V), but the combinations
    ! can be added straightfowardly by changing the array in SetUpColours
    !********************************************************************
    call CAL_setState(OM , 1, accuracyLevel = ACCURACY_HIGH)



    !****************************************
    ! Allocate memory for arrays holding
    ! the standard colours
    !*************************************** 
    stdColourMagnitude => Allocate2DSingle(1, nIndep, 1, ncol, "stdColourMagnitude")
    stdColourMagnitude = NullValue  

    stdColourMagnitudeError => Allocate2DSingle(1, nIndep, 1, ncol, "stdColourMagnitudeError")
    stdColourMagnitudeError = NullValue  

    !***************************************************
    ! Calculate the standard magnitudes and colours
    ! NB- can only use U, B and V  
    !***************************************************
    standardMagnitude => Allocate2DSingle(1, nIndep, 1, nCol, "standardMagnitude")
    standardMagnitude = NullValue

    standardMagnitudeError => Allocate2DSingle(1, nIndep, 1, nCol, "standardMagnitudeError")
    standardMagnitudeError = NullValue    

    stdMagExist = .false.
    noColours = 0

    !*****************************************************************
    ! Calculate the standard colours and magnitudes in reverse order 
    ! to preserve the priority
    ! Store the different standard magnitudes and errors for each source
    ! that have been computed for each colour combination (then average
    ! them after and calculate the colours  
    !*****************************************************************

    nCount => Allocate2DInt32(1, nCol, 1, nCol, "nCount") 
 
    stdMagnitudeStore => Allocate3DSingle(1, nIndep, 1, nCol, 1, nCol,   "stdMagnitudeStore")
    stdMagnitudeErrorStore => Allocate3DSingle(1, nIndep, 1, nCol, 1, nCol,   "stdMagnitudeErrorStore")
   
    stdMagnitudeStore = NullValue
    stdMagnitudeErrorStore = NullValue


    nCount = 0
    stdString1 = ""
    stdColourString = ""
    noColours = 0  ! The number of instrumental colours

    call addcomment(tableSet, "STANDARD MAGNITUDES- COLUMN FILTER CODES-")
   
    do i = 3, 1, -1 ! Only first 3 colours are optical

        iColumn = standardColour(i)%mag1
        jColumn = standardColour(i)%mag2

     
        if (ANY(magexist(1 : nIndep, iColumn : iColumn)) .and.  &
            ANY(magexist(1 : nIndep, jColumn : jColumn))) then

            noColours = noColours + 1
            colourString(noColours) = standardColour(i)%label1
            colourErrorString(noColours) = standardColour(i)%elabel
            stdString(noColours) = standardColour(i)%flabel

            if (noColours .eq. 1) then
                stdString1 = trim(standardColour(i)%flabel)
                stdColourString = trim(standardColour(i)%label1)
                    
            else
                label = trim(stdString1)
                stdstring1 = trim(label) // "," //  trim(standardColour(i)%flabel)

                label = trim(stdColourString)
              
                stdColourstring = trim(label) // "," //  trim(standardColour(i)%label1)
            end if

!write(*, *) "Line 651 nocolors=", nocolours, " stdstring1=", trim(stdstring1), " stdcolourstring=", trim(stdcolourstring)
!write(*, *) "line 652 standardcolour=", standardcolour(i)            


            stdString(noColours) =  trim(standardColour(i)%label1)
            stdMagExist(iColumn) = .true.
            stdMagExist(jColumn) = .true.
            
            do k = 1, nIndep
                if (mag_table(k, iColumn) .ne. NullValue .and. &
                    mag_table(k, jColumn) .ne. NullValue) then

                    !************************************************
                    ! Calculate the standard magnitudes and colour
                    !************************************************
                    call CalculateStandardColour(filt(iColumn), filt(jColumn),        &
                        dble(mag_table(k, iColumn)),  dble(magerr_table(k, iColumn)), &
                        dble(mag_table(k, jColumn)),  dble(magerr_table(k, jColumn)), &
                        stdMagnitude1, stdMagnitude1Error,                            &
                        stdMagnitude2, stdMagnitude2Error,                            &
                        stdColour, stdColourError)
                    !**************************************************************
                    ! Set nCount to 1 to record that this colour was used to
                    ! compute a standard colour
                    !************************************************************** 
                    nCount(iColumn, noColours) = 1
                    nCount(jColumn, noColours) = 1
                
                    stdMagnitudeStore(k, iColumn, noColours) = stdMagnitude1
                    stdMagnitudeErrorStore(k, iColumn, noColours) = stdMagnitude1Error 
                    stdMagnitudeStore(k, jColumn, noColours) = stdMagnitude2
                    stdMagnitudeErrorStore(k, jColumn, noColours) = stdMagnitude2Error
                    stdColourMagnitude(k, noColours) = stdColour
                    stdColourMagnitudeError(k, noColours) = stdColourError
    
                end if
            end do 

        end if
    end do
  
    deallocate(magexist, stat=allocErr)
    call CheckDeallocation(allocErr, "magexist")

   
    magnitude => Allocate2dSingle(1, nIndep, 1, nFilt, "magnitude")
    magnitude = NullValue
    magnitudeError => Allocate2dSingle(1, nIndep, 1, nFilt, "magnitudeError")
    magnitudeError = NullValue
     

    !******************************************************
    ! Add the standard magnitude and its error to the table
    ! 
    !******************************************************
    n = 0
   
    call addcomment(tableSet, "The Standard-magnitude table consists of standard magnitudes computed")
    

    call addcomment(tableSet, "using the following filter combinations")
    

    do i = 4 , nFilt - 1 ! NB NO STANDARDS FOR "WHITE OR UV FILTERS"
        stdString1 = trim(filtid(i)) // " - standard magnitude computed from "
        n = 0               
        if (stdMagExist(i))  then
   

            do j = 1, noColours
                if (nCount(i, j) .gt. 0) then
                    n = n + 1
                    magnitude(:, n) = NullValue
                    magnitudeError(:, n) = NULLValue
      
              
               
                    do k = 1, nIndep
                        magnitude(k, n) = stdMagnitudeStore(k, i, j)
                        magnitudeError(k, n) = stdMagnitudeErrorStore(k, i, j)
                    end do

                 

!                     write(comment, *) "Filter ", trim(filtid(i))
!                     call addcomment(tableset, trim(comment))
                         
                    comment = trim(stdstring1)
                    
                    stdstring1 = trim(comment) // ", " // trim(stdstring(j))                 
                    
                    write(comment, *) n, " ", trim(stdstring(j))  
                    call addcomment(tableset, trim(comment))
     

                end if
!write(*, *) "Line 709 stdstring1=", trim(stdstring1) 

            end do
        end if

        if (n .gt. 0) then
!write(*, *) "Line 713 stdstring1=", trim(stdstring1) 
     
              
             call addcomment(tableset, trim(stdstring1))
     
                 
            ! Add all the computed standard magnitudes for this filter to the table
            call AddVectorColumn(tab=tableSet, data=magnitude, ndata=nIndep, ncols=n, storeindex=RASortIndex, &
                  heading=trim(filtid(i)) // "_STD", units="mag", comment=trim(stdstring1), nullvalue=nullValue)

            call AddVectorColumn(tab=tableSet, data=magnitudeError, ndata=nIndep, ncols=n, storeindex=RASortIndex, &
                heading="ERR_" // trim(filtid(i)) // "_STD", units="mag", comment=trim(stdstring1), nullvalue=nullValue)

            ! Recall how standard magnitudes were computed
            
        end if 

    end do
 
    string = "All filters"

    !********************************************
    ! Add the standard colours to the table
    !********************************************
    string = "All filters"
    ! Add the standard mags to the table

    call AddVectorColumn(tab=tableSet, data=stdColourMagnitude, ndata=nIndep, ncols=noColours, storeindex=RASortIndex, &
            heading="STD_COL", units="mag", comment= "Standard colours (" // trim(stdcolourstring) // ")", nullvalue=nullValue)


    call AddVectorColumn(tab=tableSet, data=stdColourMagnitudeError, ndata=nIndep, ncols=noColours, storeindex=RASortIndex, &
            heading="STD_COL_ERR", units="mag", comment="Standard colour errors (" // trim(stdcolourstring) // ")", &
              nullvalue=nullValue)
 

    deallocate(magColour, magColourError, standardMagnitude, &
               standardMagnitudeError, stdColourMagnitude,   &
               stdColourMagnitudeerror, stat=allocErr)
    call CheckDeallocation(allocErr, "stdColourMagnitude")
    

    !**********************************************************
    ! Compute the fluxes
    !*********************************************************
!    do i = 1, nIndep    
!        omFitData(i)%filter = filtid(1 : 7)
!    end do

!    omFitdata%noFilters = noFluxFilters
  
    !**********************************************************
    ! Add the Vega Fluxes
    ! Use the flux conversion constants to compute fluxes
    !***********************************************************
!    call AddVegaFluxes(tableset, corrRate_table, corrRateErr_table, nIndep, rasortIndex, nofilters, filterindex, filtid)



    !**********************************************************
    ! Add the AB Fluxes
    ! Use the flux conversion constants to compute fluxes
    !***********************************************************
!    call AddABMagsandFluxes(tableset, corrRate_table, corrRateErr_table, nIndep, rasortIndex, &
!         nofilters, filterindex, filtid, filterString)





    !**********************************************************
    ! Add the White dwarf Fluxes
    ! Use the flux conversion constants to compute fluxes
    !***********************************************************
!    call AddWhiteDwarfFluxes(tableset, corrRate_table, corrRateErr_table, nIndep, rasortIndex, nofilters, filterindex, filtid)

    !**********************************************************
    ! Add the Agn Fluxes
    ! Use the flux conversion constants to compute fluxes
    !***********************************************************
!    call AddAgnFluxes(tableset, corrRate_table, corrRateErr_table, nIndep, rasortIndex, nofilters, filterindex, filtid)



!    call ComputeSourceFluxes(nIndep, omFitData, raSortIndex, tableSet)

 
   
    !*************************************
    ! Add the major-axes for each filter 
    !*************************************
!    do j = 1 , noFilters
 
!       i = filterIndex(j)
!       label = trim(filtid(i)) // "_MAJ"       
!       call Add2Real32IndexColumn(tableSet, label, "arcsecs", "Computed from moments", &
!                       nIndep, majorTable, i, RASortIndex, nullValue)
!    end do



    !*********************************************
    ! Add the major-axes for each filter
    !*********************************************

    call AddVectorColumn(tableSet, majorTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "MAJ_AXIS", "arcsecs", "Major-axis (" // trim(filterString) // ")", nullValue)


    !*********************************************
    ! Add the mean major-axes 
    !*********************************************
    call AddMeanValTableColumn(tableSet, majorTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "MEAN_MAJ_AXIS", "arcsecs", "Mean major-axis (" // trim(filterString) //")", nullValue)
    source%semia = majorTable(:, 1)


    deallocate(majorTable, stat=allocErr)
    call CheckDeallocation(allocErr, "majorTable")
 
    !*********************************************
    ! Add the minor-axes for each filter
    !*********************************************
    call AddVectorColumn(tableSet, minorTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "MIN_AXIS", "arcsecs",  "Minor-axis (" // trim(filterString) // ")", nullValue)


    !*********************************************
    ! Add the mean minor-axes 
    !*********************************************
    call AddMeanValTableColumn(tableSet, minorTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "MEAN_MIN_AXIS", "arcsecs", "Mean minor-axis (" // trim(filterString) //")", nullValue)
    source%semia = minorTable(:, 1)


    deallocate(minorTable, stat=allocErr)
    call CheckDeallocation(allocErr, "minorTable")

    !*********************************************
    ! Add the position-angle for each filter
    !*********************************************
    call AddVectorColumn(tableSet, paTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "POS_ANGLE", "degrees",  "Position-angle (" // trim(filterString) // ")", nullValue)

    !*********************************************
    ! Add the mean position-angle 
    !*********************************************
    call AddMeanValTableColumn(tableSet, paTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "MEAN_POS_ANGLE", "arcsecs",  "Mean position-angle " // trim(filterString), nullValue)

    source%theta = paTable(:, 1)


    ! Add the flags to the table 
 
    ! Loop through each filter and set the flags for each source
    do i = 1, noFilters, -1


        k = filterIndex(i) 
 
        !*************************************************************
        ! Read-out streaks
        !************************************************************
        call ReadOutStreaks(nIndep, corrRate_Table(:, k), &
                source%xpos, source%exposure, source%extensionFlag, qFlagTable(:, k))
 
 
        !*************************************************************
        ! Stars in smoke rings
        !************************************************************
        call SmokeRings(nIndep, corrRate_Table(:, k), &
                source%xpos, source%ypos, source%exposure, source%extensionFlag, qFlagTable(:, k))
 
        call BrightSourceProximityFlags(nIndep, corrRate_Table(:, k), &
                source%xpos, source%ypos, source%exposure, qFlagTable(:, k))
 





        source%corrCountrate = corrRate_Table(:, k)

        source%qualityFlag = qFlagTable(: , k)


        !*************************************************************
        ! Star diffraction spikes
        !************************************************************
        call StarSpikes(nIndep, source)


        !*************************************************************
        ! Sources near an image edge
        !************************************************************
        call EdgeSources(nIndep, source)


        qFlagTable(:, k) = source%qualityFlag


    end do


    
    call addcomment(tableset, "QFLAG- bits set (value refers to integer value, colour refers to region colour)")
    call addcomment(tableset, "if QFLAG=0 then region colour will be either")
    call addcomment(tableset, "green (point-source) or cyan (extended source)")
    call addcomment(tableset, "0=Bad pixel (1, red)")
    call addcomment(tableset, "1=Source on possible read-out streak (2, magenta)")
    call addcomment(tableset, "2=Source on possible smoke-ring (4, yellow)")
    call addcomment(tableset, "3=Source on possible star diffraction-spike (8, white)")
    call addcomment(tableset, "4=Source likely affected by mod-8 pattern (16, black)")
    call addcomment(tableset, "5=Source near bright source (32, black)")
    call addcomment(tableset, "6=Source within central enhanced region (64, red)")
    call addcomment(tableset, "EG Bad pixel + read-out streak, bits 0 and 1 set, integer value =3 colour=red")
  

    call addcomment(tableset, "CFLAG- bits Source confusion flag (value refers to integer value, color refers to region colour")
    call addcomment(tableset, "0 (point-source only) one or more sources lie within a radius of 6 to 12 unbinned pixels (1)")
    call addcomment(tableset, "1 (point-source only) one or more sources lie within a radius of 6 unbinned pixels (2)")
    call addcomment(tableset, "2 Source near an image edge (4, blue)")
    call addcomment(tableset, "3 point-source lies partly, or entirely, within an extended-source region (8)")


    
    call AddVectorColumn(tableSet, qFlagTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "QFLAG",  " ", "Quality flag (" // trim(filterString) // ")",  int(int32NullValue, int8))
 

    where(qFlagTable .eq. int32NullValue) qFlagTable = int(-2, int8)

 
    !************************************************************************
    ! Clear bits 1 (read-out streaks) ,2 (smoke-rings)  and 3 (star-spikes)
    !************************************************************************ 
    qflagTable = ibclr(qflagTable, 1)
    qflagTable = ibclr(qflagTable, 2)
    qflagTable = ibclr(qflagTable, 3)
    

    !*************************************************************
    !  Find the maximum source-counts
    !*************************************************************
    do i = 1, nIndep
        qFlagTable(i, 1) = qflagTable(i, filterIndex(1))
        do j = 2, noFilters
    
            k = filterIndex(j)
            
            if (qFlagTable(i, k) .gt. qFlagTable(i, 1)) qFlagTable(i, 1) = qFlagTable(i, k)
                    
        end do
    end do 


    where(qFlagTable(:, 1) .lt. -1) qFlagTable(:, 1) = int32NullValue
   
    do i = 1, nIndep
        source(i)%corrCountRate = -1.0
        do j = 1, noFilters
 
            k = filterIndex(j)
            if (corrRate_table(i, k) .gt. 0.0 .and. corrRate_table(i, k) .lt. 1000.0) then
                source(i)%corrCountRate = max(source(i)%corrCountRate, corrRate_table(i, k))
            end if
        end do
    end do 

!   write(*, *) "Line 977"
    source(1 : nindep)%qualityFlag = qflagTable(1 : nIndep, 1)
   
!   write(*, *) "Line 980"

 
    !*******************************************************************
    ! Quality flags for combined source-list
    !*******************************************************************
    do i = 1, nIndep
        do j = 1, 3 
            source(i)%qualityFlag = ibclr(source(i)%qualityFlag, j)
            
        end do

!        do j = 5, 5 
!            source(i)%qualityFlag = ibclr(source(i)%qualityFlag, j)            
!        end do

        if ( source(i)%qualityFlag .eq. int32NullValue) source(i)%qualityFlag = 0  
    end do

!source%qualityFlag = 0

 

    !***********************************************************************
    ! Read-out streaks
    !**********************************************************************
    call ReadOutStreaks(nIndep, noFilters, filterIndex, corrRate_Table, &
        source%xpos, source%exposure, source%extensionFlag, source%qualityFlag)
 

    !***********************************************************************
    ! Sources near to a bright source
    !**********************************************************************


    call BrightSourceProximityFlags(nIndep, corrRate_Table(:, k), &
                source%xpos, source%ypos, source%exposure, source%qualityFlag)
 


    !***********************************************************************
    ! Sources within a smoke ring
    !**********************************************************************


    call SmokeRings(nIndep, noFilters, filterIndex, corrRate_Table, &
        source%xpos, source%ypos, source%exposure, source%extensionFlag, source%qualityFlag)


    !***********************************************************************
    ! Sources on a star diffraction spike
    !**********************************************************************
    call StarSpikes(nIndep, source)
 
    !***********************************************************************
    ! Sources on an edge
    !**********************************************************************

    call EdgeSources(nIndep, source)



    qFlagTable(:, 1) = source%qualityFlag



    call Add2Int8IndexColumn(tableSet, "QFLAG_COMB", "", "Quality flag using combined source-list", &
               nIndep, qFlagTable, 1, RASortIndex, int32nullValue)
 
 


    call AddVectorColumn(tableSet, cFlagTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "CFLAG", " ", "Confusion flag (" // trim(filterString) // ")",  int(int32NullValue, int8))


    cFlagTable(:, 1) = cflagTable(:, filterIndex(1))
    where(cFlagTable(:, 1) .eq. int32NullValue) cFlagTable(:, 1) = int(-1, int8)
  
    do i = 1, nIndep         
        do j = 2, noFilters    
            k = filterIndex(j)
            if (cFlagTable(i, k) .gt. int32NullValue) then
                if (cFlagTable(i, k) .gt. cFlagTable(i, 1)) cFlagTable(i, 1) = cFlagTable(i, k)
            end if
        end do
    end do 

    where(cFlagTable(:, 1) .lt. int(0, int8)) cFlagTable(:, 1) = int32NullValue

    call Add2Int8IndexColumn(tableSet, "CFLAG_MAX", "", "Maximum source confusion flag", &
               nIndep, cFlagTable, 1, RASortIndex, int32nullValue)

 
    call AddVectorColumn(tableSet, eFlagTable, nIndep, noFilters, filterIndex, RASortIndex, &
            "EFLAG",  "", "Extension flag (" // trim(filterString) // ")", &
            int(int32NullValue, int8))


    eFlagTable(:, 1) = eflagTable(:, filterIndex(1))
    where(eFlagTable(:, 1) .eq. int32NullValue) eFlagTable(:, 1) = int(-1, int8)
  

    do i = 1, nIndep         
        do j = 2, noFilters    
            k = filterIndex(j)
            if (eFlagTable(i, k) .gt. int32NullValue) then
                if (eFlagTable(i, k) .gt. eFlagTable(i, 1)) eFlagTable(i, 1) = eFlagTable(i, k)
            end if
        end do
    end do 

    where(eFlagTable(:, 1) .lt. int(0, int8)) eFlagTable(:, 1) = int32NullValue


    call Add2Int8IndexColumn(tableSet, "EFLAG_MAX", "", "Maximum source extension flag", &
               nIndep, eFlagTable, 1, RASortIndex, int32nullValue)

 

    deallocate(corrRate_Table, corrRateErr_table)
    nullify(corrRate_Table, corrRateErr_table)


    deallocate(eflagTable,stat=allocErr)
    call CheckDeallocation(allocErr, "eflagTable")
  
 

    !****************************************************************
    ! Astrometry corection- optional 
    !****************************************************************
    if (booleanParameter("usecat")) then


        call DoAstrometryCorrection(nIndep,                  &
                                  raSortIndex,               &
                                  nFilt,                     &
                                  source%rightascension,     &
                                  source%declination,        &
                                  mag_table,                 &
                                  magerr_table,              &
                                  source%poserr,             &
                                  raz, decz,                 &
                                  tableSet,                  &
                                  oswFile)


            comment = "Computed from corrected RAs and Decs"

   
        else

             comment = "Computed from uncorrected RAs and Decs"

    end if
   
    position = columnNumber(tableSet, "POSERR") + 1
 



    !*************************************************** 
    ! Compute the galactic longitude and latitude 
    ! from the averaged RA and dec
    !***************************************************

    galong => AllocateDouble(1, nSources, "galong")
    galat => AllocateDouble(1, nSources, "galat")

    do i = 1, nIndep
        call sla_EQGAL(source(i)%rightascension, source(i)%declination, &
                galong(i), galat(i)) 
    end do  

    galong = galong * radiansToDegrees
    

    call AddReal64IndexColumn(tableSet, "LII", "degrees", &
                 trim(comment), nIndep, galong, RASortIndex, nullValue, position=position)
    deallocate(galong, stat = allocErr)
    call CheckDeallocation(allocErr, "galong")

    galat = galat * radiansToDegrees
    position = position + 1
    call AddReal64IndexColumn(tableSet, "BII", "degrees", &
             trim(comment), nIndep, galat, RASortIndex, nullValue, position=position)
    deallocate(galat, stat = allocErr)
    call CheckDeallocation(allocErr, "galat")





    call release(tableSet)
    call release(outDataSet)



    deallocate(mag_table,magerr_table, magnitude, &
               stdMagnitudeStore, stdMagnitudeErrorStore, &
               magnitudeError, stat = allocErr)

    deallocate(source, paTable, stat=allocErr)
    deallocate(oswFile, RASortIndex, &
               nrows, stat=allocErr)
    call CheckDeallocation(allocErr, "nrows")
    
!    deallocate(omFitData)
!    nullify(omfitdata)
end subroutine omsrclistcomb

!**********************************************************
! Add a source ID column to each of the osw source lists
! telling us which source went to which entry in the
! observation source list 
!**********************************************************
subroutine AddSourceIds(nosw, oswFile, matchedNumber, nrows, raSortIndex)
    integer, intent(in) :: nosw
    character(*), pointer :: oswFile(:)
    integer, pointer :: matchedNumber(:) 
    integer, pointer :: nrows(:)
    integer, pointer :: raSortIndex(:)

    integer(int32), pointer :: sourceID(:) => null()
   
    integer :: nStart, i, j, k
    type(dataSetT) :: oswDataSet
    type(TableT) :: tableSet
 
 
    nStart = 1
 
    write(errstr, *) "Adding source ids"
    call message(AppMsg, Verbose)
          
    sourceID => allocateInt32(1, maxval(nrows), "sourceId")
  
    k = 0 ! Source number in sorted uncombined list

!do i = 1, size(raSortIndex)
!write(*, *) i, " sort index=", rasortindex(i)
!end do
    do i = 1 , nosw ! Loop through each input source-list

        oswDataSet = DataSet(oswFile(i), modify)
        tableSet = table(oswDataSet, 0)
   
        do j = nStart, nStart + nrows(i) - 1    ! Add ID number to each source

            k = k + 1

!write(*, *) "k=", k, " matchednumber k=", matchedNumber(k)
            sourceID(j - nStart + 1) = rasortIndex(rasortindex(matchednumber(k)))

        end do

        call AddInt32Column(tableSet, sourceId, nRows(i), &
               "SRC_ID", "", "Source number in observation source-list file")
      
        call release(tableSet)
        call release(oswDataSet)
               
        nStart = nStart + nrows(i)

    end do

    deallocate(sourceId)

end subroutine  AddSourceIds

!****************************************************
! CalculateStandardColour - Given the two filters, two
! instrumental magnitudes and colours, calculates the 
! standard colours. If the instrumental colours are outside
! the validity range the values are set to nullValue.
!*****************************************************
subroutine CalculateStandardColour(filterID1, filterID2, &
           instrumentalMagnitude1, instrumentalMagnitude1Error, & 
           instrumentalMagnitude2, instrumentalMagnitude2Error, &
           standardMagnitude1, standardMagnitude1Error, & 
           standardMagnitude2, standardMagnitude2Error, &
           standardColour, standardColourError)
  
    integer(int32) :: filterID1, filterID2
    real(double), intent(in) :: instrumentalMagnitude1, instrumentalMagnitude1Error
    real(double), intent(in) :: instrumentalMagnitude2, instrumentalMagnitude2Error
            
    real(double), intent(out) :: standardMagnitude1, standardMagnitude2                     
    real(double), intent(out) :: standardColour
    real(double), intent(out) :: standardColourError
    real(double), intent(out) :: standardMagnitude1Error
    real(double), intent(out) :: standardMagnitude2Error  
    real(double)    :: magnitudeRange
    type(OmColorTransformatorT)     :: colourTrans
    type(IntervalT)                 :: colourValInt


    
    colourTrans = CAL_omGetColorTransformator(filterID1, filterID2)
    colourValInt = CAL_omColorTransValidityRange(colourTrans)  


    !*******************************************************
    ! Check that the instrumental magnitudes lie within
    ! the allowed range.
    !*******************************************************
    magnitudeRange = instrumentalMagnitude1 - instrumentalMagnitude2

    if (magnitudeRange .gt. colourValInt%lower .and. &
        magnitudeRange .lt. colourValInt%upper)  then    ! Range okay                       
       
        !****************************************************
        ! NB The first magnitude in the following call is
        ! transformed into the standard system
        !****************************************************
        call CAL_omColorTransform(filterId1, filterID2, &
                  instrumentalMagnitude1, instrumentalMagnitude2, &
                  standardColour,  standardMagnitude2, &
                  standardColourError, standardMagnitude2Error)


        standardMagnitude1Error = AddErrors(standardMagnitude2Error,  standardColourError) 

!***************************************************************************
! Temporary (18.2.2002) Because the cal routine does not yet calculate the 
! standard magnitude errors properly, just set the error equal to
! the instrumental magnitude error
!***************************************************************************  

        standardMagnitude1Error = instrumentalMagnitude1Error
        standardMagnitude2Error = instrumentalMagnitude2Error
        standardMagnitude1 = standardMagnitude2 + standardColour

    end if 
                                           
end subroutine CalculateStandardColour

!******************************************
! CreateObsSourceList
! create the output observation source list
!******************************************
subroutine CreateObsSourceList(noOswFiles, oswFile, outDataSet, raPoint, decPoint, exposure)
    type(DataSetT), intent(out) :: outDataSet
    integer, intent(in) :: noOswFiles
    character(*), intent(in), dimension(1 : noOswFiles) :: oswFile
    real(double), intent(out) :: raPoint, decPoint
    real(single), intent(out) :: exposure
    character(40) :: startTimeString, endTimeString
    type(DataSetT) :: oswDataSet
    integer i, k
    character(40) :: object, observer, obsID, startObs, endObs

    outDataSet = DataSet(stringParameter("outset"), create)
  
    call setAttribute(outDataSet, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")
    call setAttribute(outDataSet, "HDUCLAS1", "SRCLIST", "File contains a source list")
    call setAttribute(outDataSet, "FILTER", "n/a", "Not applicable")
    call setAttribute(outDataSet, "CONTENT", "SRCLIST", "File content")



    !************************************************
    ! 21.2.2002 - Add the new attributes for SCR 84
    !************************************************
    call OAL_addCommonAttributes(outDataSet)



    !************************************************
    ! Find the first start time and the last end time
    !************************************************
    oswDataSet = DataSet(oswFile(1), read)

    startTimeString = stringAttribute(oswDataSet, "DATE-OBS")
    endTimeString   = stringAttribute(oswDataSet, "DATE-END")

    if (hasAttribute(oswDataSet, "OBS_ID")) then
        obsID   = stringAttribute(oswDataSet, "OBS_ID")
    else
        obsId = "00000"
    end if
    if (hasAttribute(oswDataSet, "RA_PNT")) then
        raPoint = real64Attribute(oswDataSet, "RA_PNT")
        decPoint = real64Attribute(oswDataSet, "DEC_PNT")
    else
        raPoint = 0.0_double
        decPoint = 0.0_double
    end if

    object   = trim(stringAttribute(oswDataSet, "OBJECT"))

    if (hasAttribute(oswDataSet, "OBSERVER")) then
 
        observer   = trim(stringAttribute(oswDataSet, "OBSERVER"))
    else
        observer = "MAX"
    end if
    exposure   = real32Attribute(oswDataSet, "EXPOSURE")
    
  
    call release(oswDataSet)

    do i = 2, noOswFiles
 
        oswDataSet = DataSet(oswFile(i), read)
        startObs = stringAttribute(oswDataSet, "DATE-OBS")
        endObs   = stringAttribute(oswDataSet, "DATE-END")
        if (trim(stringAttribute(oswDataSet, "OBJECT")) .ne. object) call fatal("wrong object")
        


        !******************************************************
        ! Need the camera pointing direction to do the 
        ! astrometry correction
        !***************************************************** 
        if (hasAttribute(oswDataSet, "RA_PNT")) then
            raPoint = raPoint + real64Attribute(oswDataSet, "RA_PNT")
            decPoint = decPoint + real64Attribute(oswDataSet, "DEC_PNT")
         


            k = CompareDates(startTimeString, startObs)
        
            if (k .lt. 0) startTimeString = startObs

            k = CompareDates(endTimeString, endObs)
        
            if (k .gt. 0) endTimeString = endObs
        end if  
        call release(oswDataSet)  
    end do

    raPoint = raPoint / dble(noOswFiles)
    decPoint = decPoint / dble(noOswFiles)


    call setAttribute(outDataSet, "OBS_ID", obsID, "Observation identifier")
    call setAttribute(outDataSet, "OBJECT", object, "Object")
    call setAttribute(outDataSet, "OBSERVER", observer, "Observer")
 
    call setAttribute(outDataSet, "DATE-OBS", startTimeString, "Start time of observations")
    call setAttribute(outDataSet, "DATE-END", endTimeString, "End time of observations")
    call setAttribute(outDataSet, "RA_PNT", raPoint, "degrees", "Averaged RA pointing direction")
    call setAttribute(outDataSet, "DEC_PNT", decPoint, "degrees", "Averaged Dec pointing direction")
       

end subroutine CreateObsSourceList
!*******************************************************
! AddErrors
!
!*******************************************************
real(kind=double) function AddErrors(error1, error2)
    real(double), intent(in) :: error1, error2
 
    AddErrors = dsqrt(error1 * error1  + error2 * error2)

end function AddErrors
!******************************************************
! subroutine SetUpInstrumentalColourTable
! Initialises the array with the possible colours
! NB Filter codes are V=V, U=U, B=B, White=W, Grism2=G,
! UVW1=L, UVM2=M, UVW2=S, Grism1=H.
!******************************************************
subroutine SetUpInstrumentalColourTable(Colour)

    type(ColourT), intent(out), dimension(:) :: Colour
   
    !***************** 
    ! These are OPT-OPT
    !*****************
    Colour(1) = ColourT(4, 5, "U_B", "U-B", "ERR_U_B", "")
    Colour(2) = ColourT(5, 6, "B_V", "B-V", "ERR_B_V", "")
    Colour(3) = ColourT(4, 6, "U_V", "U-V", "ERR_U_V", "")

    !****************
    ! These are UV-UV
    !****************
    Colour(4) = ColourT(1, 3, "S_L", "S-L", "ERR_S_L", "")
    Colour(5) = ColourT(1, 2, "S_M", "S-M", "ERR_S_M", "")
    Colour(6) = ColourT(2, 3, "M_L", "M-L", "ERR_M_L", "")

    !*****************
    ! These are UV-OPT
    !*****************
    Colour(7)  = ColourT(1, 4, "S_U", "S-U", "ERR_S_U", "")
    Colour(8)  = ColourT(1, 5, "S_B", "S-B", "ERR_S_B", "")
    Colour(9)  = ColourT(1, 6, "S_V", "S-V", "ERR_S_W", "")
    Colour(10) = ColourT(2, 4, "M_U", "M-U", "ERR_M_U", "")
    Colour(11) = ColourT(2, 5, "M_B", "M-B", "ERR_M_B", "")
    Colour(12) = ColourT(2, 6, "M_V", "M-V", "ERR_M_V", "")
    Colour(13) = ColourT(3, 4, "L_U", "L-U", "ERR_L_U", "")
    Colour(14) = ColourT(3, 5, "L_B", "L-B", "ERR_L_B", "")
    Colour(15) = ColourT(3, 6, "L_V", "L-V", "ERR_L_V", "")

end subroutine SetUpInstrumentalColourTable
!******************************************************
! subroutine SetUpStandardColourTable
! Initialises the array with the possible colours
! Letters refer to letter codes for filters
! ie V=V, U=U, B=B, W=White, G=Grism2, X=UVW1,
!   Y=UVM2, Z=UVW2 & H=Grism1
!******************************************************
subroutine SetUpStandardColourTable(Colour)

    type(ColourT), intent(out), dimension(:) :: Colour
     
    Colour(1) = ColourT(4, 5, "U_B", "U-B", "ERR_U_B", "_U_B")
    Colour(2) = ColourT(5, 6, "B_V", "B-V", "ERR_B_V", "_B_V")
    Colour(3) = ColourT(4, 6, "U_V", "U-V", "ERR_U_V", "_U_V")
    Colour(4) = ColourT(3, 4, "L_U", "L-U", "ERR_L_U", "_L_U")
    Colour(5) = ColourT(3, 5, "L_B", "L-B", "ERR_L_B", "_L_B")
    Colour(6) = ColourT(3, 6, "L_V", "L-V", "ERR_L_V", "_L_V")
    Colour(7) = ColourT(1, 5, "S_B", "S-B", "ERR_S_B", "_S_B")
    Colour(8) = ColourT(1, 6, "S_V", "S-V", "ERR_S_V", "_S_V")


end subroutine SetUpStandardColourTable
 
!*******************************************************
! CompareDates - Compares the first date with the second.
! Returns -1 if the second is less than the first,
! 0 if they are equal and 1 if greater.
!*******************************************************
integer function CompareDates(dateString1, dateString2)
    character(*), intent(in) :: dateString1, dateString2
    integer :: time1(6), time2(6)        ! Stores the time from the date string
    character(1) char    ! Used for date and time extraction 
    integer i

    read(dateString1, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time1(1), char, time1(2), char, time1(3), char, time1(4), &
         char, time1(5), char, time1(6)

    read(dateString2, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time2(1), char, time2(2), char, time2(3), char, time2(4), &
         char, time2(5), char, time2(6)
  
    do i = 1, 6

        CompareDates = time2(i) - time1(i)
        if (CompareDates .ne. 0) return
 
    end do

    CompareDates = 0 ! Must be equal

end function CompareDates
!************************************************************
! CompareOffsets
!
!************************************************************
subroutine CompareOffsets(fileName)
    character(*) :: fileName(:)
   
    real(single), pointer, dimension(:) ::  deltaRa, deltaRaError, &
                          deltaDec, deltadecError, &
                          deltaTheta, deltaThetaError
    real(single) :: deltaRamean, deltaRastandardDeviation
    real(single) :: deltaDecmean, deltaDecstandardDeviation
    real(single) :: deltaThetamean, deltaThetastandardDeviation
    real(single) :: minRaOffset, maxRaOffset
    real(single) :: minDecOffset, maxDecOffset

    character(4), pointer :: filter(:)

    logical, pointer :: status(:) => null()
    integer :: i, j, k
    integer(int8) :: noFilters
    character(4), pointer :: filterList(:), exposureNo(:)
    integer(int32), pointer :: sourceIndex(:)
    integer(int32)  :: noFiles
  
    noFiles = size(fileName)
    allocate(status(1 : noFiles))
    write(errstr, *) "Checking astrometric offsets in each source-list file"
    call message(AppMsg, Verbose)
      


    deltaRa => allocateSingle(1, noFiles, "deltaRa")
    deltaRaError => allocateSingle(1, noFiles, "deltaRa")
    deltaDec => allocateSingle(1, noFiles, "deltaRa")
    deltaDEcError => allocateSingle(1, noFiles, "deltaRa")
    deltaTheta => allocateSingle(1, noFiles, "deltaTheta")
    deltaThetaError => allocateSingle(1, noFiles, "deltaTheta")
    sourceIndex => allocateInt32(1, noFiles, "sourceIndex")

    allocate(filter(1 : noFiles))
    allocate(filterList(1 : noFiles))
    allocate(exposureNo(1 : noFiles))
    

    k = 0
    minRaOffset = 1e6
    maxRaOffset = -1e6
    minDecOffset = 1e6
    maxDecOffset = -1e6

    noFilters = 0
iLoop : do i = 1, noFiles
 
        call GetOffsets(trim(fileName(i)), deltaRa(i), deltaRaError(i), &
                        deltaDec(i), deltadecError(i), &
                        deltaTheta(i), deltaThetaError(i), filter(i), exposureNo(i), status(i))
        if (status(i)) then
            k = k  + 1
            deltaRa(k) = deltaRa(i)
            deltaRaError(k) = deltaRaError(i)
            deltaDec(k) = deltaDec(i)
            deltaDecError(k) = deltaDecError(i)
            deltaTheta(k) = deltaTheta(i)
            deltaThetaError(k) = deltaThetaError(i)
            minRaOffset = min(minRaOffset,  deltaRa(k))
            maxRaOffset = max(maxRaOffset,  deltaRa(k))
            minDecOffset = min(minDecOffset,  deltaDec(k))
            maxDecOffset = max(maxDecOffset,  deltaDec(k))
            filter(k) = filter(i) 
            exposureNo(k) = exposureNo(i)
            do j = 1, noFilters 
                if (trim(filter(i)) == trim(filterList(j))) cycle iLoop
            end do
            noFilters = noFilters + 1
            filterList(noFilters) = trim(filter(i))
                    
        end if 
       
    end do iLoop 
    do i = 1, k
        
        do j = 1, noFilters 
            if (trim(filter(i)) == trim(filterList(j))) then
                sourceIndex(i) = j
                exit
            end if
        end do 
    end do 
    if (k > 1) then


        call ComputeAverage(deltaRa, deltaRaError, k, deltaRamean, deltaRastandardDeviation)
        call ComputeAverage(deltaDec, deltaDecError, k, deltaDecmean, deltaDecstandardDeviation)
        call ComputeAverage(deltaTheta, deltaThetaError, k, deltaThetamean, deltaThetastandardDeviation)
  
       
        write(errstr, '(a, i3)') "No of computed astrometric offsets=", k
        call message(AppMsg, Verbose)
    
        write(errstr, '(a, f6.2, a, f6.2)') "Mean RA offset=", deltaRamean, "+/-", deltaRastandardDeviation
        call message(AppMsg, Verbose)
        write(errstr, '(a, f6.2, a, f6.2)') "Mean dec offset=", deltaDecmean, "+/-", deltaDecstandardDeviation
        call message(AppMsg, Verbose)
        write(errstr, '(a, f6.2, a, f6.2, a)') "Ra offset range=", minRaOffset, "-",  maxRaOffset, " arcsecs"
        call message(AppMsg, Verbose)
        write(errstr, '(a, f6.2, a, f6.2, a)') "Dec offset range=", minDecOffset, "-",  maxDecOffset, " arcsecs"
        call message(AppMsg, Verbose)
  
    end if

    deallocate(deltaRa, deltaRaError, deltaDec, deltaDecError, deltaTheta, deltaThetaError) 
    deallocate(filter, filterList, exposureNo, sourceIndex, status)
end subroutine CompareOffsets    
!*****************************************************
! Get the offsets from teh source-list file        
!*****************************************************
subroutine GetOffsets(sourceList, deltaRa, deltaRaError, &
                        deltaDec, deltadecError, &
                        deltaTheta, deltaThetaError, filter, exposureNo, status)
    character(*), intent(in) :: sourceList
    real(single), intent(out) ::  deltaRa, deltaRaError, &
                          deltaDec, deltadecError, &
                          deltaTheta, deltaThetaError
    character(*), intent(inout) :: filter, exposureNo
    character(20) :: text
    logical, intent(out) :: status
    type(dataSetT) :: dataSetH
    type(tableT)   :: sourceTable
      dataSetH = dataSet(sourcelist, read)

    filter = trim(stringAttribute(dataSetH, "FILTER")) !(11 : 13) 
       
    text = trim(stringAttribute(dataSetH, "EXP_ID")) !(11 : 13) 
    exposureNo = text(11 : 13)
 
 
    sourceTable = table(dataSetH, "SRCLIST")
    if (.not. hasAttribute(sourceTable, "RAOFFSET")) then
        status = .false.
        call release(sourceTable) 
        call release(dataSetH)
 
        return
    end if 
    
    if (.not. booleanAttribute(sourceTable, "POSCOROK")) then
        status = .false.
        call release(sourceTable) 
        call release(dataSetH)
 
        return
    end if 
    

    deltaRa = real32Attribute(sourceTable, "RAOFFSET")
    deltaRaError = real32Attribute(sourceTable, "RAOFFERR")
    deltaDec = real32Attribute(sourceTable, "DEOFFSET")
    deltaDecError = real32Attribute(sourceTable, "DEOFFERR")
    deltaTheta = real32Attribute(sourceTable, "ROT_CORR")
    deltaThetaError = real32Attribute(sourceTable, "ROT_ERR")
    


    call release(sourceTable) 
    call release(dataSetH)
    status = .true.
   
end subroutine GetOffsets

!**********************************
! ComputeAverage
! Calculates the weighted average and
! standard error
!**********************************
subroutine ComputeAverageMagnitude(magnitude, error, size, mean, standardError)
    integer, intent(in) :: size        
    real(single), intent(in), dimension(1 : size) :: magnitude, error    
    real(single), intent(out) :: mean, standardError
    real(double) :: weight
    real(double) :: sum1, sum2, sum3
    real(double) :: d, difference
    logical :: useWeights
    integer i, nData
    real(double), pointer       :: flux(:) => null()
    real(double), pointer       :: fluxError(:) => null()
  
    if (size .eq. 1) then
        mean = magnitude(1)
        standardError = error(1)
        return
    end if


    flux => AllocateDouble(1, size, "flux")
    fluxError => AllocateDouble(1, size, "fluxError")

    nData = 0
    do i = 1, size
        if (magnitude(i) .ge. 0.0 .and. magnitude(i) .lt. 100.0) then
            nData = nData + 1 
            flux(nData) = dble(10.0 ** (-0.4 * magnitude(i)))                 
            fluxError(nData) = dble(flux(ndata) * error(i))
        end if
    end do   
    
    
    if (nData .le. 1) then
        mean = magnitude(1)
        standardError = error(1)
        deallocate(flux)
        deallocate(fluxError)
        return
    end if


    useWeights = .true.
    weight = 1.0_double

   
    sum1 = 0.0_double
    sum2 = 0.0_double
    sum3 = 0.0_double
 

    do i = 1, nData
       
        if (fluxError(i) .le. 0.0) then
            useWeights = .false.
            sum1 = dble(ndata)
            sum3 = dble(ndata)
            exit
        end if
    end do 

    
    do i = 1, nData
       
       if (useWeights) then
            weight = 1.0_double / (fluxError(i) * fluxError(i))
            sum1 = sum1 + weight
            sum2 = sum2 + flux(i) * weight
            sum3 = sum3 + weight * weight
        else
            sum2 = sum2 + flux(i)
          
        end if

    end do

    mean = sum2 / sum1  ! The mean

    sum2 = 0.0_double

    do i = 1, nData
       
   
        difference = flux(i) - mean       

        if (useWeights) then
         
            sum2 = sum2 + fluxError(i) * fluxError(i)
        else
            sum2 = sum2 + difference * difference
        end if 
    end do

    if (useWeights) then
        d = sum1 - sum3 / sum1
    else
        d = dble(ndata - 1)
    end if



    standardError = sqrt(sum2) / d 
    
    standardError = max(0.01, 1.085735 * standardError / mean)
  


    deallocate(flux, fluxError)
    

  
    mean = -2.5 * log10(mean)

end subroutine ComputeAverageMagnitude


!*************************************************
!
! Subroutine GetMagnitudes
!*************************************************
subroutine GetMagnitudes(nFilt, nIndep, &
            insMagnitudeStore, insMagnitudeErrorStore, & 
            magnitude, magnitudeError, nullValue)


    integer(int32), intent(in) :: nFilt, nIndep
    real(single), pointer       :: insMagnitudeStore(:, :)   
    real(single), pointer       :: insMagnitudeErrorStore(:, :)
    real(single)       :: magnitude(:)   
    real(single)       :: magnitudeError(:)   
    real(single)                :: nullValue              ! Null
    integer(int32) :: i, k, count, maxCount, index1

   
    maxCount = 0
    index1 = 1
   

    do i = 1 , nFilt

   
        count = 0 
        do k = 1, nIndep
            if (insMagnitudeStore(k, i) > 0.0 .and.  &
                insMagnitudeStore(k, i) < 30.0) count = count + 1
 
                      
        end do
        if (count > maxCount) then
            maxCount = count
            index1 = i
                       
        end if  
          
     
   
    end do
 
    magnitude = NullValue
    magnitudeError = NULLValue
      
    do i = 1, nIndep
        magnitude(i) = insMagnitudeStore(i, index1)
        magnitudeError(i) = insMagnitudeErrorStore(i, index1)
    end do

end subroutine GetMagnitudes


!subroutine CorrectSourceListFile(fileName, deltaRamean,  deltaRaStandardDeviation, &
!                                 deltaDecmean, deltaDecStandardDeviation, &           
!                                 deltaThetaMean, deltaThetaStandardDeviation)
     
!    character(*), intent(in) :: fileName
!    real(single), intent(in) :: deltaRamean,  deltaRaStandardDeviation, &
!                                 deltaDecmean, deltaDecStandardDeviation, &           
!                                 deltaThetaMean, deltaThetaStandardDeviation
!    real(double) :: raPoint, decPoint
!    integer(int32) :: nrows
!    real(double), pointer :: ra(:), dec(:)
!    real(double) :: coeffs(6)
!    type(DataSetT) :: dataSetH
!    type(tableT) :: tableSet

!    write(errstr, *) "Adding RA/dec corrected columns in file ", &
!                trim(fileName), " using the computed mean offsets"
!    call message(AppMsg, Verbose)
  
       
!    dataSetH = DataSet(fileName, read)
    
!    tableSet = table(dataSetH, 0)
     
!    raPoint = real64Attribute(dataSetH, "RA_PNT")
!    decPoint = real64Attribute(dataSetH, "DEC_PNT")
          

!    nrows = numberOfRows(tableSet)

!    ra => allocateDouble(1, nrows, "ra")
!    dec => allocateDouble(1, nrows, "dec")
  
!    call ReadReal64Column(tableSet, "RA", nrows, 0, ra)
!    call ReadReal64Column(tableSet, "DEC", nrows, 0, dec)

              


!    call release(tableSet)
!    call release(dataSetH)
!    coeffs(1) =  deltaRamean * degreesToRadians / 3600.0_double
!    coeffs(2) =  deltaDecmean * degreesToRadians / 3600.0_double
!    coeffs(3) = 1.0_double
!    coeffs(4) = 1.0_double
!    coeffs(5) = 0.0_double
!    coeffs(6) =  deltaThetamean * degreesToRadians
!    call CorrectSourcePositions(nrows, ra, dec, &
!                                coeffs, raPoint, decPoint)

    

!    call AddCorrectionColumns(fileName, nrows, ra, dec, &
!                              deltaRamean,  deltaRaStandardDeviation, &
!                              deltaDecmean, deltaDecStandardDeviation, &           
!                              deltaThetaMean, &
!                              deltaThetaStandardDeviation, 0, dble(0.0))
    
!    deallocate(ra, dec)
        
!end subroutine CorrectSourceListFile
!*******************************************
! Function MissionTime - Returns the number
! of seconds from the start of the mission
! to the observation time. Added by CB because
! the call to CAL_GetBoresiteMatrix needs
! the time in seconds from the launch.
!******************************************
real(double) function  MissionTime(primaryTable)
    type(DataSetT), intent(in)   :: primaryTable

    character(len = 19)  :: dateObs         
    character(len = 19)  :: dateEnd        
    real(kind = double)  :: JulianDate0, JulianDate1, JulianDate2
    character(len = 19)  :: missionReferenceTime = "1998-01-01T00:00:00"

    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the start of the observations.
    dateObs = stringAttribute(primaryTable,  "DATE-OBS")    
    JulianDate1 = ModifiedJulianDate(dateObs)

    ! Compute the Julian date of the end of the observations.
    dateEnd = stringAttribute(primaryTable,  "DATE-END")   
    JulianDate2 = ModifiedJulianDate(dateEnd)

    ! The required time is the difference between the average of the 
    ! start and end times and the mission reference time, in seconds.
    MissionTime = ((JulianDate1 + JulianDate2) / 2.0d0 - JulianDate0) * 86400.0d0

end function MissionTime



!**********************************************************************
!
! Add the flags to the table
!*********************************************************************
subroutine AddFlagsToTable(tableSet, noFilters, filterIndex, nIndep, &
         flagTable, tableName, raSortIndex, intNullValue)


    type(tableT), intent(inout) :: tableSet
    integer, intent(in) :: noFilters
    integer :: filterIndex(:)
    integer, intent(in) :: nIndep
    integer(int8),  pointer, dimension(:, :)   :: flagTable
    character(*), intent(in) :: tableName
    integer, pointer :: raSortIndex(:)
    integer, intent(in) :: intNullValue    
    integer :: i, j, k, index
    character(14), pointer :: flagCode(:) => null()

    allocate(flagCode(1 : nIndep))

    flagCode = "       "
    


    do i = 1, nIndep
        flagCode = "--------"
        do j = 1, noFilters
            index = filterIndex(j)
            k = (j - 1) * 2 + 1
            if (flagTable(i, index) .ne. intNullValue) then
              
                write(flagCode(i)(k : k + 1), '(i2)') fLagTable(i, index)
        
              
            end if
!write(*, *) i, ") ", flagTable(i, index), " code= ", flagCode(i)
        end do
    end do 

    call AddCharIndexColumn(tableSet, tableName, " ", " ", nIndep, RASortIndex, flagCode)



    deallocate(flagCode)
    nullify(flagCode)

end subroutine AddFlagsToTable






!********************************************************************************
!
!*******************************************************************************


subroutine AddInstrumentalMags(tableSet, noFilters, filterIndex, &
             nIndep, mag_table, magerr_table, RASortIndex, nullValue)
    type(TableT), intent(inout) :: tableSet
    integer, intent(in) :: noFilters
    integer, intent(in) :: filterIndex(:)
    integer, intent(in) :: nIndep
    real(single), pointer :: mag_table(:, :), magerr_table(:, :)
    integer(int32), pointer :: RAsortIndex(:)
    real(single), intent(in) :: nullValue
    



!    do j  = 1 , noFilters

!        i = filterIndex(j)


!        label = trim(filtid(i)) // "_INS"
!        call Add2Real32IndexColumn(tableSet, label, "magnitudes", "Instrumental magnitudes", &
!                           nIndep, mag_table, i, RASortIndex, nullValue)

!        label = "ERR_" // trim(filtid(i)) // "_INS"
!        call Add2Real32IndexColumn(tableSet, label, "magnitudes", "Instrumental", &
!               nIndep, magerr_table, i, RASortIndex, nullValue)



!    end do


    call AddVectorColumn(tableSet, mag_table, nIndep, noFilters, filterIndex, RASortIndex, &
            "INS_MAG", "mags", "Instrumental", nullValue)

    call AddVectorColumn(tableSet, magerr_table, nIndep, noFilters, filterIndex, RASortIndex, &
            "INS_MAG_ERR", "mags", "Instrumental", nullValue)





end subroutine AddInstrumentalMags


!***********************************************************************
! This subroutine tries to fit the template-spectra to each source
! The best-fit template is used to compute the fluxes for the spectrum
! If no fit is done then fluxes are computed using the 
! zero-magnitude flux conversion constants from the ccf
!
!**********************************************************************


!subroutine ComputeSourceFluxes(nRows, omFitData, raSortIndex, tableSet)
!    integer, intent(in) :: nRows    
!    type(sourceData), pointer:: omFitData(:)
!    integer :: raSortIndex(:)
!    type(TableT), intent(inout) :: tableSet
  
  
!    integer :: noFits
!    type(fittedData) :: fit                                                      

!    logical(boolean) :: fitted
!    real(double) :: rmsMinimum
!    integer(int16) :: templateNo, noOfDegreesOfFreedom
!    integer :: i
 

   
 
    
    ! Load the stellar templates
!    call LoadTemplateData(templateData)

!    open(unit=10, status="UNKNOWN", file="fittedspectra.dat")
!    write(10, '(a12, a10, a10, a5)') "Source no", & 
!           "Temp No", "Rms Res", "MU"

!    call pgbegin(0, "spectra.ps" // '/VC', 1, 1)
  
       
!    noFits = 0


!    fit%flux = 0.0
!    fit%fluxErr = 0.0

    !******************************************************************
    ! Loop through all the sources and try a fit a template spectrum
    !******************************************************************
!    do i = 1, nRows
 
!        omFitData(i)%flux(1 : 7) = nullValue
!        omFitData(i)%fluxErr(1 : 7) = nullValue
  
        ! Try to fit a spectrum

!        call FitSpectra(i, omFitData(i), templateData, fitted, fit)
!        if (fitted) then ! Successful fit to template spectrum
!            call updateLogFile(i, rmsMinimum, templateNo,  noOfDegreesOfFreedom)
            
            ! Plot the fit
!            call PlotFit(i, fit)


!            omFitData(i)%flux = fit%flux
!            omFitData(i)%fluxErr = fit%fluxErr
!            noFits = noFits + 1   
         
!        end if
       
!    end do

!    call pgclos
!    write(10, '(a, i5, a, i5, a)') "Total no of fitted spectra=", noFits, " (no of sources=", nRows, ")"
      
!    write(errstr, '(a, i5, a, i5, a)') "Total no of fitted spectra=", noFits, " (no of sources=", nRows, ")"
   
!    call message(AppMsg, Verbose)
!    write(10, *) errstr          
!    close(10)
   



    ! Add the new flux columns
!    call AddTemplateFluxColumns(nRows, omFitData, raSortIndex, tableSet )


!end subroutine ComputeSourceFluxes

!****************************************************************
!
!
!**************************************************************
subroutine DoAstrometryCorrection(nIndep,       &
                                  raSortIndex,  &
                                  nFilt,        &
                                  ra,           &
                                  dec,          &
                                  mag_table,    &
                                  magerr_table, &
                                  poserr,       &
                                  raz, decz,    &
                                  tableSet,     &  
                                  oswFile)

    integer, intent(in) :: nIndep, raSortIndex(:), nfilt
    real(double), dimension(:) :: ra, dec
    real(single), pointer, dimension(:, :) :: mag_table, magerr_table
    real(single) :: poserr(:)
    real(double), intent(in)    :: raz, decz
    character(*), pointer       :: oswFile(:)             ! filenames
    type(TableT), intent(inout) :: tableSet
    real(double), pointer       :: raCorr(:)=>Null()      ! Correctewd RA
    real(double), pointer       :: decCorr(:)=>Null()     ! Corrected ded
    real(single), pointer :: magnitude(:, :) => null()
    real(single), pointer :: magnitudeError(:, :) => null()
    real(single), pointer :: magTemp(:) => null()    
    real(single) :: maxRADecErr = 1.0, maxRmsResidual = 1.5
    logical :: status = .false.
    logical :: okay = .false.

  
    type(CorrectionData) :: corrData    
    type(CatalogueData), pointer :: catData(:) => null()   

    integer :: i 

    magTemp => AllocateSingle(1, nIndep, "magTemp")
    
    magnitude => Allocate2dSingle(1, nIndep, 1, nFilt, "magnitude")
    magnitudeError => Allocate2dSingle(1, nIndep, 1, nFilt, "magnitudeError")

    magnitude = NullValue
 

    !*****************************************************************
    ! Get the magnitudes for the astrometry correction
    !*****************************************************************
    call GetMagnitudes(nFilt, nIndep, & 
                        mag_table, magerr_table, & 
                        magnitude(:, 1), magnitudeError(:, 1), nullValue)

    raCorr => AllocateDouble(1, nIndep, "raCorr")
    decCorr => AllocateDouble(1, nIndep, "decCorr")
  

    ! NB ra & dec have to be in radians            
    do i = 1, nIndep
        raCorr(i)  = ra(raSortIndex(i)) * radiansToDegrees
        decCorr(i) = dec(raSortIndex(i)) * radiansToDegrees
    end do

      
    magTemp => allocateSingle(1, nIndep, "magTemp")
    magTemp = magnitude(:, 1)



    call CorrectAstrometry(nIndep, ra, dec, &
                           posErr, magTemp, raz, decz,  oswFile(1), &
                           status, corrData, catData)     

      
    deallocate(magTemp)
    nullify(magTemp)
 

    if (status) then  ! The astrometry correction was successful
      
        maxRADecErr = realParameter("maxradecerr")
        maxRmsResidual = realParameter("maxrmsres")


 

        ! Check if fitted parameters are okay 
        if (fitOkay(corrData, maxRADecErr, maxRmsResidual)) then
            okay = .true.

            do i = 1, nIndep
                raCorr(i)  = ra(raSortIndex(i)) * radiansToDegrees
                decCorr(i) = dec(raSortIndex(i)) * radiansToDegrees
            end do
 

           


!            if (booleanParameter("plotoffsets")) then
!                call pgbegin(0, "omsrclistcomboffsets.ps/VC", 1, 1)
     
!                call PlotSkyMap(nIndep, real(ra * radiansToDegrees), real(dec * radiansToDegrees), &
!                               size(catData), real(catData%ra * radiansToDegrees), &
!                       real(catData%dec * radiansToDegrees), "OM uncorrected source positions")

!                call PlotSkyMap(nIndep, real(raCorr), real(decCorr), &
!                               size(catData), real(catData%ra * radiansToDegrees), &
!                   real(catData%dec * radiansToDegrees), "OM corrected source positions")

!            end if 
        end if
      
    end if
   


    call AddCorrectionColumns2(tableSet, &
                 nIndep, raCorr, decCorr, corrData)

    

    ra = raCorr
    dec = decCorr
    deallocate(raCorr, decCorr, magnitude, magnitudeError)


    ! Add the POSCOROK attribute
   
  
   
    call setAttribute(tableSet, "POSCOROK", okay, &
                              "ASTROMETRY CORRECTION OKAY")
    call setAttribute(tableSet, "SYSERROR", 0.5, &
                              "Systematic catalogue position errors")



    ! Check the computed offsets in each source-list file
    call CompareOffsets(oswFile)


end subroutine DoAstrometryCorrection

!********************************************************
!
!
!*******************************************************
subroutine LoadSourceData(oswFile, source, nStart)
    character(*), intent(in) :: oswFile
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: nStart
    type(TableT) :: tableSet
    type(DataSetT) :: oswDataSet
    integer :: nRows
    integer(int32) :: windowx0, windowy0, xbin, ybin, filterIdInt
    real(double) :: plateScale
    character(4) :: omFilter
    real(single) :: exposure

    oswDataSet = DataSet(oswFile, read)
    tableSet = table(oswDataSet, 0)
    nrows = numberOfRows(tableSet)
    source(nStart + 1 : nStart + nrows)%exposure = real32Attribute(oswDataSet, "EXPOSURE") ! The exposure time (secs)
   
    ! Get the coordiantes
    call ReadReal32Column(tableSet, "XPOS", nrows, nStart, source%xpos)
    call ReadReal32Column(tableSet, "YPOS", nrows, nStart, source%ypos)

    ! Convert the pixel coordinates to full-frame coordinates (0.5- 2048.5, 0.5-2048.5)

    omFilter = stringAttribute(oswDataSet, "FILTER")
    !  Get the filter Id integer value
    filterIdInt = filterIdToInt(omFilter)

    call CAL_setState(OM, filterId = filterIdInt)


    ! Get the Plate scale (arcsecs/pixel)
    call CAL_omGetPlateScale(0.55_double, plateScale)


    call GetWindowParameters(oswDataSet, windowX0, windowY0, xbin, ybin)
    call ConvertCoordsToFullFrame(source%xpos, source%ypos, nStart + 1, nStart + nRows, oswDataSet)   

    exposure = real32Attribute(oswDataSet, "EXPOSURE")   
    source(nStart + 1 : nStart + nrows)%exposure = exposure
   
  
    ! Get the position error
    call ReadReal32Column(tableSet, "POSERR", nrows, nStart, source%poserr)

!if (any(source(nstart + 1: nstart+nrows)%poserr .lt. 0.000001)) then
!call fatal("bug at line 3037")
!end if

!write(*, *) "platescale=", platescale, " binx=", binx
    !***********************************************************************
    ! Convert position to arc secs
    !********************************************************************** 
    source(nStart + 1 : nstart + nrows)%poserr = &
               real(dble(source(nstart + 1 : nstart + nrows)%poserr) * plateScale * dble(xbin))


!if (any(source(nstart + 1: nstart+nrows)%poserr .lt. 0.000001)) then
!call fatal("bug at line 3049")
!end if





    ! Get the Right-Ascension and declination
    call ReadReal64Column(tableSet, "RA", nrows, nStart, source%rightAscension)
    call ReadReal64Column(tableSet, "DEC", nrows, nStart, source%declination)

    ! Get the source major and minor axes and position angle
    call ReadReal32Column(tableSet, "FWHM_MAJ", nrows, nStart, source%semia)
    call ReadReal32Column(tableSet, "FWHM_MIN", nrows, nStart, source%semib)
    call ReadReal32Column(tableSet, "PA", nrows, nStart, source%theta)
    
    ! Get the significance
    call ReadReal32Column(tableSet, "SIGNIFICANCE", nrows, nStart, source%significance)

    ! Get the instrumental magnitudes
    call ReadReal32Column(tableSet, "MAG", nrows, nStart, source%magnitude)
    call ReadReal32Column(tableSet, "MAGERR", nrows, nStart, source%magnitudeerr)

    ! Get the rates for spectral-fitting anf for checking for sources near read-out streaks
    call ReadReal32Column(tableSet, "CORR_RATE", nrows, nStart, source%corrCountRate)
    call ReadReal32Column(tableSet, "CORR_RATE_ERR", nrows, nStart, source%corrCountRateErr)

    call ReadReal32Column(tableSet, "CORR_BACKGROUND_RATE", nrows, nStart, source%corrBackgroundCountRate)
!    call ReadReal32Column(tableSet, "CORR_BACKGROUND_RATE_ERR", nrows, nStart, source%corrBackgroundCountRateErr)


    
    
    ! Get the source flags
    call ReadInt8Column(tableSet, "QFLAG", nrows, nStart, source%qualityflag)

    

    call ReadInt8Column(tableSet, "CFLAG", nrows, nStart, source%confusionflag)
    call ReadInt8Column(tableSet, "EFLAG", nrows, nStart, source%extensionflag)
  

    call release(oswDataSet)
       
end subroutine LoadSourceData


!***************************************************
! Function SourceMatch
! Tries to match a catalogue star to a given source
!
!***************************************************
logical function SourceMatch1(il, iu, source, catSource, &
                           foundfile, tolerance, minRadius, matched, index)
 
    integer, intent(in) :: il, iu
    type(sourceParameters), intent(in) :: source
    type(sourceParameters), pointer :: catSource(:)
    logical :: foundFile(:)
    real(single), intent(in) :: tolerance
    real(single), intent(out) :: minRadius
    logical :: matched(:)  
    integer, intent(out) :: index     
    real(double) :: distance
    real(double) ::  error1, error2
    real(single) :: error
    real(double), save :: delta = (20.0_double / 3600.0_double) * degreesToRadians 

    integer(int32) i, kl, ku, jStart, n

    if (il .gt. iu) then
        SourceMatch1 = .false.
        return
    end if
!
!write(*, *) "Matching source with ra", source%rightAscension
!do i = il, iu
!write(*, *) i, " cat ra=", catSource(i)%rightAscension
!end do   
    jStart = 1   



    n = iu - il + 1
 
    ! Bracket the catalogue stars that might match the star

    kl = Locate(n, catSource(il : iu)%rightAscension, source%rightAscension - delta, jStart) + il - 1
    ku = Locate(n, catSource(il : iu)%rightAscension, source%rightascension + delta, jStart) + il - 1

!write(*, *) "Line 2431 kl=", kl, " ku=", ku
                   
    minRadius = 10000.0
    
    sourcematch1 = .false.


    if (source%extensionFlag .eq. 0) then
        error1 = max(source%poserr, 1.0) ! was 2
    else
        error1 = max(source%poserr, 3.0)
    end if 
    

!    kl = il
!    ku = iu

    do i = kl, ku
        if (matched(i)) cycle

   
        if (foundfile(catsource(i)%imin)) cycle ! Don't match sources from the same file
      
        if (abs(catsource(i)%declination - source%declination) .gt. delta) cycle
        
        distance = ArcLength(source%rightAscension, source%declination, catSource(i)%rightascension, catSource(i)%declination)

        if (catSource(i)%extensionFlag .eq. 0) then
            error2 = max(catSource(i)%poserr, 1.0) ! was 1.0
        else
            error2 = max(catSource(i)%poserr, 3.0)
        end if 
   


        error = nsigma * AddErrors(error1, error2)

              
        error = max(tolerance, error)


        if (distance .le. error) then  

            if(distance .lt. minRadius) then
                minRadius = distance
                SourceMatch1 = .true.
                index = i
            end if
                
        end if

    end do
!call fatal("stopped")

    if (sourceMatch1) then
        matched(index) = .true.

        foundfile(catsource(index)%imin) = .true.

    end if

end function SourceMatch1   


!**************************************************************
!
!*************************************************************
subroutine SortDataRAOrder(source, filter, n)
    type(sourceParameters) :: source(:)
    character(*), pointer :: filter(:)
    integer, intent(in) :: n 
    type(sourceParameters), pointer :: tempSource(:) => null()
    character(len=10), pointer :: tempFilter(:) => null()
    integer(int32), pointer :: raSortIndex(:) => null()
    integer :: i
    

    rasortindex => allocateint32(1, n, "rasortindex")

    allocate(tempSource(1 : n)) ! For storing the combined source-list
    allocate(tempFilter(1 : n)) ! For storing the combined source-list


    tempsource = source

    call IndexSort(tempsource%rightAscension, RASortIndex, n)
    tempsource = source
    tempFilter = filter
    do i = 1, n
          tempsource(i) = source(rasortindex(i))
          tempFilter(i) = filter(rasortindex(i))
 
    end do

    source = tempSource
    filter = tempFilter
    deallocate(rasortindex, tempSource, tempFilter)
    nullify(rasortindex, tempSource, tempFilter)


end subroutine SortDataRAOrder


!**********************************************************
! Add the Vega Fluxes
! Use the flux conversion constants to compute fluxes
!***********************************************************
subroutine AddVegaFluxes(srcListTable, counts, countserr, nrows, rasortIndex, nfilters, filterindex, filter)
    type(TableT) :: srcListTable
    real(single), pointer, dimension(:, :) :: counts, countserr
    integer, intent(in) :: nrows
    integer, pointer:: rasortindex(:) 
    integer, intent(in) :: nfilters, filterIndex(:)
    character(*) :: filter(:)
    real(single), pointer :: fluxTable(:, :) => null()
    real(single), pointer :: fluxErrTable(:, :) => null()
    integer :: i, j, k, position
    character(100) :: comment
    ! Flux conversions taken from sas watch-out pages 
    real(single), parameter :: VegafluxConvert(7) = (/4.31e-15, 1.78e-15, 4.74e-16, 1.60e-16, 1.20e-16, 2.19e-16, -1.0/)
    type(dataSetT) dataSetH

    dataSetH = parent(srclistTable)

    fluxTable => allocate2dsingle(1, nrows, 1, 7, "fluxtable")
    fluxErrTable => allocate2dsingle(1, nrows, 1, 7, "fluxerrtable")
    fluxTable = nullValue
    fluxErrTable = nullValue
  
    call addcomment(srclisttable, "VEGA_FLUX - Computed fluxes using Vega calibration for the following filters")
   


    do i = 1, nfilters
        write(comment, *) i, "=", trim(filter(filterIndex(i)))  
        call addcomment(srclisttable, trim(comment))
   

    end do

    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .lt. 999.0) then 
                if (j .eq. 7) then
                    fluxTable(i, k) = nullValue
                    fluxErrTable(i, k) = nullValue
                else
                    fluxTable(i, k) =  counts(i, j) * VegafluxConvert(j)            
                    fluxerrTable(i, k) = countsErr(i, j) *  VegafluxConvert(j)        
                end if  
            end if 

        end do
    end do  

    if (hasColumn(srclisttable, "STD_COLS_ERR")) then    
        position = columnNumber(srclisttable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srclisttable)
    end if
 


    call AddVectorColumn(tab=srcListTable, data=fluxtable, ndata=nrows, &
            ncols=nfilters, storeindex=RASortIndex, &
            heading="VEGA_FLUX", units="ergs/s/cm**2/A",  &
            comment="Fluxes based on Vega flux calibration ", &
            nullvalue=nullValue, position=position)

    call AddVectorColumn(tab=srcListTable, data=fluxerrtable, ndata=nrows, &
             ncols=nfilters, storeindex=RASortIndex, &
            heading="VEGA_FLUX_ERR", units="ergs/s/cm**2/A", &
            comment="Flux errors based on Vega flux calibration ", & 
            nullvalue=nullValue, position=position + 1)

    deallocate(fluxtable, fluxerrtable)
    nullify(fluxtable, fluxerrtable)

end subroutine AddVegaFluxes






!**********************************************************
! Add the AB magnitudes and Fluxes 
! Use the flux conversion constants to compute fluxes
!***********************************************************
subroutine AddABMagsandFluxes(srcListTable, counts, countserr, nrows, &
           rasortIndex, nfilters, filterindex, filter, filterString)
    type(TableT) :: srcListTable
    real(single), pointer, dimension(:, :) :: counts, countserr
    integer, intent(in) :: nrows
    integer, pointer:: rasortindex(:) 
    integer, intent(in) :: nfilters, filterIndex(:)
    character(*) :: filter(:)
    character(*), intent(in) :: filterString
    real(single), pointer :: fluxTable(:, :) => null()
    real(single), pointer :: fluxErrTable(:, :) => null()
    integer :: i, j, k, position
    character(100) :: comment
    
    !*********************************************************************
    ! The following constants are taken from Antonio's e-mail of 21.07.05
    !*********************************************************************

    real(single), parameter :: ABfluxConvert(7) = (/2.50e-16, 1.25e-16, &
             1.94e-16, 4.82e-16, 2.21e-15, 5.70e-15, -1.0/)

    real(single), parameter :: zeropoint(7) = (/16.5719, 17.4120, 18.5662, 19.1890, 19.0809, 17.9230, -1.0/)

    type(dataSetT) dataSetH

    dataSetH = parent(srclistTable)

    fluxTable => allocate2dsingle(1, nrows, 1, 7, "fluxtable")
    fluxErrTable => allocate2dsingle(1, nrows, 1, 7, "fluxerrtable")

    fluxTable = nullValue
    fluxErrTable = nullValue
  
    call addcomment(srclisttable, &
         "AB Magnitudes and fluxes (AB_MAG, AB_MAG_ERR, AB_FLUX, AB_FLUX_ERR), computed for the following filters")
   
    do i = 1, nfilters
        write(comment, *) i, "=", trim(filter(filterIndex(i))), " zero-point=", zeroPoint(i), &
                    " flux conversion constant=", abfluxconvert(i)  
        call addcomment(srclisttable, trim(comment))
   
    end do





    !*************************************************************
    ! Add the AB magnitudes
    !************************************************************

    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .lt. 999.0) then 
                if (j .eq. 7) then
                    fluxTable(i, k) = nullValue
                    fluxErrTable(i, k) = nullValue
                else
                    fluxTable(i, k) =  zeropoint(j) - 2.5 * log10(counts(i, j))            
                    fluxerrTable(i, k) = countsErr(i, j) *  ABfluxConvert(j)        
                    fluxerrTable(i, k) = max(0.01, 1.085735 * countsErr(i, j) / counts(i, j))

                end if

            end if 

        end do
    end do  

    if (hasColumn(srclisttable, "STD_COLS_ERR")) then    
        position = columnNumber(srclisttable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srclisttable)
    end if
 


    call AddVectorColumn(tab=srcListTable, data=fluxtable, ndata=nrows, &
            ncols=nfilters, storeindex=RASortIndex, &
            heading="AB_MAG", units="mag",  &
            comment="AB mags (" // trim(filterString) // ")", & 
            nullvalue=nullValue, position=position)

    call AddVectorColumn(tab=srcListTable, data=fluxerrtable, ndata=nrows, &
             ncols=nfilters, storeindex=RASortIndex, &
            heading="AB_MAG_ERR", units="mag", &
            comment="AB mag errors (" // trim(filterString) // ")", & 
            nullvalue=nullValue, position=position + 1)


    !*************************************************************
    ! Add the AB fluxes
    !************************************************************
    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .lt. 999.0) then 
                if (j .eq. 7) then
                    fluxTable(i, k) = nullValue
                    fluxErrTable(i, k) = nullValue
                else
                    fluxTable(i, k) =  counts(i, j) * ABfluxConvert(j)            
                    fluxerrTable(i, k) = countsErr(i, j) *  ABfluxConvert(j)        
                
                end if
            end if 

        end do
    end do  

    call AddVectorColumn(tab=srcListTable, data=fluxtable, ndata=nrows, &
            ncols=nfilters, storeindex=RASortIndex, &
            heading="AB_FLUX", units="ergs/s/cm**2/A",  &
            comment="AB fluxes (" // trim(filterString) // ")", & 
            nullvalue=nullValue, position=position + 2)

    call AddVectorColumn(tab=srcListTable, data=fluxerrtable, ndata=nrows, &
             ncols=nfilters, storeindex=RASortIndex, &
            heading="AB_FLUX_ERR", units="ergs/s/cm**2/A", &
            comment="AB flux errors (" // trim(filterString) // ")", & 
            nullvalue=nullValue, position=position + 3)

    deallocate(fluxtable, fluxerrtable)
    nullify(fluxtable, fluxerrtable)

end subroutine AddABMagsandFluxes




!**********************************************************
! Add the White dwarf  Fluxes
! Use the flux conversion constants to compute fluxes
!***********************************************************
subroutine AddWhiteDwarfFluxes(srcListTable, counts, countserr, nrows, rasortIndex, nfilters, filterindex, filter)
    type(TableT) :: srcListTable
    real(single), pointer, dimension(:, :) :: counts, countserr
    integer, intent(in) :: nrows
    integer, pointer:: rasortindex(:) 
    integer, intent(in) :: nfilters, filterIndex(:)
    character(*) :: filter(:)
    real(single), pointer :: fluxTable(:, :) => null()
    real(single), pointer :: fluxErrTable(:, :) => null()
    integer :: i, j, k, position
    character(100) :: comment
   
    real(single), parameter :: whitedwarffluxConvert(7) = (/4.76e-15, 2.17e-15, 5.80e-16, 1.98e-16, 1.33e-16, 2.53e-16, -1.0/)

    type(dataSetT) dataSetH

    dataSetH = parent(srclistTable)


    fluxTable => allocate2dsingle(1, nrows, 1, 7, "fluxtable")
    fluxErrTable => allocate2dsingle(1, nrows, 1, 7, "fluxerrtable")
    fluxTable = nullValue
    fluxErrTable = nullValue
  
    call addcomment(srclisttable, "WHITEDWARF_FLUX - Computed fluxes using White Dwarf calibration for the following filters")
   


    do i = 1, nfilters
        write(comment, *) i, "=", trim(filter(filterIndex(i)))  
        call addcomment(srclisttable, trim(comment))
     

    end do

    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .lt. 999.0) then 
                if (j .eq. 7) then
                    fluxTable(i, k) = nullValue
                    fluxErrTable(i, k) = nullValue
                else
                    fluxTable(i, k) =  counts(i, j) * whitedwarffluxConvert(j)            
                    fluxerrTable(i, k) = countsErr(i, j) *  whitedwarffluxConvert(j)        
             

                 end if           
                  
            end if 

        end do
    end do  

    if (hasColumn(srclisttable, "STD_COLS_ERR")) then    
        position = columnNumber(srclisttable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srclisttable)
    end if
 


    call AddVectorColumn(tab=srcListTable, data=fluxtable, ndata=nrows, &
            ncols=nfilters, storeindex=RASortIndex, &
            heading="WHITEDWARF_FLUX", units="ergs/s/cm**2/A",  &
            comment="Fluxes based on White dwarf  flux calibration ", &
            nullvalue=nullValue, position=position)

    call AddVectorColumn(tab=srcListTable, data=fluxerrtable, ndata=nrows, &
             ncols=nfilters, storeindex=RASortIndex, &
            heading="WHITEDWARF_FLUX_ERR", units="ergs/s/cm**2/A", &
            comment="Flux errors based on White dwarf  flux calibration ", & 
            nullvalue=nullValue, position=position + 1)

    deallocate(fluxtable, fluxerrtable)
    nullify(fluxtable, fluxerrtable)

end subroutine AddWhiteDwarfFluxes


!**********************************************************
! Add the AGN Fluxes
! Use the flux conversion constants to compute fluxes
!***********************************************************
subroutine AddAGNFluxes(srcListTable, counts, countserr, nrows, rasortIndex, nfilters, filterindex, filter)
    type(TableT) :: srcListTable
    real(single), pointer, dimension(:, :) :: counts, countserr
    integer, intent(in) :: nrows
    integer, pointer:: rasortindex(:) 
    integer, intent(in) :: nfilters, filterIndex(:)
    character(*) :: filter(:)
    real(single), pointer :: fluxTable(:, :) => null()
    real(single), pointer :: fluxErrTable(:, :) => null()
    integer :: i, j, k, position
    character(100) :: comment
    real(single) :: VegafluxConvert(7) = (/4.871e-15, 2.103e-15, 4.585e-16, 1.889e-16, 1.541e-16, 2.385e-16, -1.0/)
    type(dataSetT) dataSetH

    dataSetH = parent(srclistTable)



    fluxTable => allocate2dsingle(1, nrows, 1, 7, "fluxtable")
    fluxErrTable => allocate2dsingle(1, nrows, 1, 7, "fluxerrtable")
    fluxTable = nullValue
    fluxErrTable = nullValue
  
    call addcomment(srclisttable, "AGN_FLUX - Computed fluxes using AGN calibration for the following filters")
 

    do i = 1, nfilters
        write(comment, *) i, "=", trim(filter(filterIndex(i)))  
        call addcomment(srclisttable, trim(comment))
   

    end do

    do i = 1, nrows
        do k = 1, nfilters
            j = filterIndex(k)            
            if (counts(i, j) .gt. 0.0 .and. counts(i, j) .lt. 999.0) then 
                fluxTable(i, k) =  counts(i, j) * VegafluxConvert(j)            
                fluxerrTable(i, k) = countsErr(i, j) *  VegafluxConvert(j)        
                  
            end if 

        end do
    end do  

    if (hasColumn(srclisttable, "STD_COLS_ERR")) then    
        position = columnNumber(srclisttable, "STD_COLS_ERR") + 1
    else
        position = numberOfColumns(srclisttable)
    end if
 


    call AddVectorColumn(tab=srcListTable, data=fluxtable, ndata=nrows, &
            ncols=nfilters, storeindex=RASortIndex, &
            heading="AGN_FLUX", units="ergs/s/cm**2/A",  &
            comment="Fluxes based on Vega flux calibration ", &
            nullvalue=nullValue, position=position)

    !******************************************************************
    ! Add the flux errors
    !*****************************************************************
    call AddVectorColumn(tab=srcListTable, data=fluxerrtable, ndata=nrows, &
             ncols=nfilters, storeindex=RASortIndex, &
            heading="AGN_FLUX_ERR", units="ergs/s/cm**2/A", &
            comment="Flux errors based on Vega flux calibration ", & 
            nullvalue=nullValue, position=position + 1)

    deallocate(fluxtable, fluxerrtable)
    nullify(fluxtable, fluxerrtable)

end subroutine AddAgnFluxes


!*********************************************************************
!
!
!********************************************************************
subroutine CreateMergedSourceList(n, source, nfiles, fileName, filter, &
                  nrows, nfilt, filterId, sourceFilter, nnew, newSource, matchedNumber)
    integer, intent(in) :: n
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: nFiles
    character(*), pointer :: fileName(:)
    character(*), pointer :: filter(:) ! List of filter for each file
    integer :: nrows(1 : nfiles)
    integer, intent(in) :: nFilt 
    character(*) :: filterId(1 : nfilt)
    character(*), pointer :: sourceFilter(:)     ! List of filters for each source
    integer, intent(out) :: nnew
    type(sourceParameters), pointer :: newSource(:)
    integer(int32), pointer :: matchedNumber(:)
    integer, pointer :: index(:, :), sourceCount(:)=> null()
    integer(int32) :: i, j, k
    type(sourceParameters), pointer ::tempSource1(:) => null()
    character(5), pointer :: tempsourceFilter1(:) => null()
    integer :: nFiltSources(1 : nfilt), maxFilt, maxFiltIndex 
    integer                     :: allocErr
  
    write(errstr, '(a)') "Creating merged source-list"
    call message(AppMsg, Verbose)
    write(errstr, '(a)') "Input source-list files"
    call message(AppMsg, Verbose)
  
    write(errstr, '(a2, a4, a13, a4, a6, a4, a20)') "No", "  ",  "no of sources",   " ",  "filter", " ", "file name"
    call message(AppMsg, Verbose)
  
    sourceCount => allocateint32(1, n, "sourcecount")
    index => allocate2dint32(1, n, 1, nfiles, "index")
   

    k = 0     
    do i = 1, nfiles
        write(errstr, '(i5, i13, a17, a4, a)') i, nrows(i), trim(filter(i)), " ", trim(fileName(i)) 
        call message(AppMsg, Verbose)
        do j = 1, nrows(i)
            k = k + 1
            source(k)%imin = i  ! Remember which file this source came from
            source(k)%imax = j  ! Remember the row number in the source list file 
            source(k)%number = k  ! Need to remember the original source number
  
        end do
       
    end do





    !**********************************
    ! 
    !*********************************
    call AligncoordinateAxes(n, source, nfiles, nrows)

    !*************************************************
    ! Sort the data into increasing RA order
    !************************************************      
    call SortDataRAOrder(source, sourceFilter, n)


    !******************************************************
    ! Find the filter that has the most sources
    !******************************************************

    nFiltSources = 0
    maxFilt = 0

    do j = 1, nfilt
        do i = 1, nfiles
            if (trim(filter(i)) .eq. trim(filterid(j))) nfiltSources(j) = nfiltSources(j) + nrows(i)
        end do
        if (nfiltSources(j) .gt. 0) then
            write(errstr, '(a, a, a, i6, a)') "Filter ", trim(filterid(j)), " has ", nfiltSources(j), " sources" 
            call message(AppMsg, Verbose)
            if (nFiltSources(j) .gt. maxFilt) then
                maxFiltIndex = j
                maxFilt = nFiltSources(j)
            end if
        end if

    end do


    write(errstr, '(a, a, a, i6, a)') "Filter ", trim(filterid(maxFiltIndex))," has the maximum number of sources (", maxFilt, ")" 
    call message(AppMsg, Verbose)
     

    !***************************************************************************
    ! Match the sources
    !***************************************************************************
    call FindIdenticalSources(n, source, nfiles, nnew, index, sourceCount)


    do i = 1, n
        matchedNumber(i) = source(i)%noSourcePixels
    end do





    allocate(newSource(1 : nnew))
    allocate(tempSource1(1 : maxval(sourceCount)))
    allocate(tempSourceFilter1(1 : maxval(sourceCount)))
  


    !***************************************
    ! Allocate memory for output arrays.
    !***************************************
    mag_table => Allocate2DSingle(1, nnew, 1, nFilt, "mag_table")
    magerr_table => Allocate2DSingle(1, nnew, 1, nFilt, "magerr_table")

    ! Count-rate tables (for spectral fitting and CCD streak quality-flag setting)
    corrRate_table => Allocate2DSingle(1, nnew, 1, nFilt, "corrrate_table")
    corrRatEerr_table => Allocate2DSingle(1, nnew, 1, nFilt, "corrrateerr_table")


!    allocate(omFitData(1 : nSources)) ! The data for spectrum-fitting
    

                
    significanceTable => Allocate2DSingle(1, nnew, 1, nFilt, "significanceTable")

    ! Ellipse paramaterisation tables
    majorTable => Allocate2DSingle(1, nnew, 1, nFilt, "majorTable")
    minorTable => Allocate2DSingle(1, nnew, 1, nFilt, "minorTable")
    paTable => Allocate2DSingle(1, nnew, 1, nFilt, "paTable")

    ! Quality flag tables
    qflagTable => Allocate2DInt8(1, nnew, 1, nFilt, "qflagTable")
    cflagTable => Allocate2DInt8(1, nnew, 1, nFilt, "cflagTable")
    eflagTable => Allocate2DInt8(1, nnew, 1, nFilt, "eflagTable")
   

    allocate(magexist(1 : nnew, 1 : nFilt), stat=allocErr)
    call CheckAllocation(allocErr, "magExist")

    mag_table = NullValue
    magerr_table = NullValue
    corrRate_table = NullValue
    corrRateErr_table = NullValue
    significanceTable = NullValue
    majorTable = NullValue
    minorTable = NullValue
    paTable = NullValue
    qflagTable = int(int32NullValue, int8)
    cflagTable = int(int32NullValue, int8)
    eflagTable = int(int32NullValue, int8)
  

    magexist = .false.

    do i = 1, nnew

        do j = 1, sourceCount(i)
            tempSource1(j) = source(index(i, j))
            tempSourceFilter1(j) = sourcefilter(index(i, j))
!            write(*, *) j, " ) x=", tempSource1(j)%xpos, " y=", tempSource1(j)%ypos
!            write(*, *) "Filter =", tempSourceFilter1(j)  
        end do
 
        !*****************************************************************************
        ! Create the counts, etc, for this source (for each filter)
        !*****************************************************************************   
        call MergeSources(i, nFilt, filterid, &
                              tempSourceFilter1, &
                              sourceCount(i), &
                              tempsource1, &
                              newSource(i))    
    end do  


    if (associated(tempSource1)) then
        deallocate(tempSource1)
        nullify(tempSource1)
    end if
   
    if (associated(tempSourceFilter1)) then
        deallocate(tempSourceFilter1)
        nullify(tempSourceFilter1)
    end if
    
    write(errstr, '(a)') "No   - source combined from files"
    call message(AppMsg, Verbose)


end subroutine CreateMergedSourceList
!********************************************************************
! Subroutine MergeSource
! n input sources, source(1 : n) are merged into 1
!
!*******************************************************************
subroutine MergeSources(i, nFilt, filterId, sourceFilter, n, source, mergedSource)
    integer, intent(in) :: i
    integer, intent(in) :: nFilt
    character(*) :: filterId(1 : nFilt), sourcefilter(:)  
    integer, intent(in) :: n
    type(sourceParameters) :: source(:)
    type(sourceParameters), intent(out):: mergedSource
    integer :: j, k, iflag, nmag, nmag1     
    real(double) :: ra_tmp, dec_tmp, poserr_tmp, weightsum, weight
    real(double) :: xi, eta
    real(single), dimension(1 : n) :: xPosTemp, yPosTemp, xyPosErrTemp
    real(single), dimension(1 : n) :: magTemp, magErrTemp,  corrRateTemp, corrRateErrTemp
    real(single) :: mean, standardDeviation
    real(single) :: significanceTemp, majorTemp, minorTemp, paTemp, exposureTemp

!integer, save :: count = 0

    !*************************************** 
    ! find the weighted mean of the position
    !***************************************      


!write(*, *) "line n=", n, " merging sources-"
!do j = 1, n
!write(*, *) j, source(j)%significance
!end do

    ra_tmp = 0.0_double
    dec_tmp = 0.0_double
    poserr_tmp = 0.0_double
    weightSum = 0.0
    majorTemp = 0.0
    minorTemp = 0.0
    paTemp = 0.0
    significanceTemp = 0.0
    exposureTemp = 0.0  
 
    if (n .eq. 1) then ! Only one source
        mergedSource = source(1)


        do j = 1 , nFilt        ! Find which filter it belongs to

            if ( trim(sourceFilter(1)) .eq. trim(filterid(j)) ) then
 
                qflagTable(i, j) = source(1)%qualityFlag 
                cflagTable(i, j) = source(1)%confusionFlag
                eflagTable(i, j) = source(1)%extensionFlag
       
                mag_table(i, j) = source(1)%magnitude

                magerr_table(i, j) = source(1)%magnitudeErr
                  
                corrRate_table(i, j) = source(1)%corrCountRate  
          
                corrRateErr_table(i, j) = source(1)%corrCountRateErr
!                omFitData(i)%counts(j) = source(1)%corrCountRate
!                omFitData(i)%countserr(j) = source(1)%corrCountRateErr

                significanceTable(i, j) = source(1)%significance

                majorTable(i, j) = source(1)%semia
          
                minorTable(i, j) = source(1)%semib
          
        
                paTable(i, j)   = source(1)%theta
         
                magexist(i, j) = .true.
            
                return ! Found the filter so can return
            end if
        end do
     

    end if


    !********************************************************************
    ! Set each bit of the quality and confusion flags to the maximum set
    !********************************************************************
    mergedSource%qualityFlag = maxval(source(1 : n)%qualityFlag)         
    mergedSource%confusionFlag = maxval(source(1 : n)%confusionFlag)

    ! Set each bit of the extension flag to the minimum set   
    mergedSource%extensionFlag = minval(source(1 : n)%extensionFlag)

    do k = 1,  n ! Average

        if (source(k)%poserr .gt. 0.0) then
            weight = 1.0_double / (source(k)%poserr * source(k)%poserr)
        else
            weight = 1.0_double
        end if    
 
        weightSum = weightSum + weight
        !*************************************************************
        ! Convert RAs and decs to tangent-plane coordinates- average
        ! the sum of these and then convert the average to RA and dec
        !*************************************************************

        call sla_ds2tp(source(k)%rightascension, source(k)%declination, raz, decz, xi, eta, iflag)

        ra_tmp = ra_tmp + xi * weight     
        dec_tmp = dec_tmp + eta * weight
        poserr_tmp = poserr_tmp + source(k)%poserr * weight
        xPosTemp(k) = source(k)%xpos        ! X-coordinate
        yPosTemp(k) = source(k)%ypos        ! Y-coordinate
        xyPosErrTemp(k) = source(k)%poserr          

        majorTemp = majorTemp + source(k)%semiA
        minorTemp = minorTemp + source(k)%semiB
        paTemp = paTemp + source(k)%theta
        significanceTemp = significanceTemp + source(k)%significance
        exposureTemp = exposureTemp + source(k)%exposure
     

    end do

   
    ! The mean x coordinate
    call ComputeAverage(xPosTemp, xyposErrTemp, n, mergedsource%xpos, standardDeviation)        
    ! The mean y coordinate
    call ComputeAverage(yPosTemp, xyPosErrTemp, n, mergedsource%ypos, standardDeviation)        


    mergedsource%poserr = poserr_tmp / weightSum

    ! Compute the average RA and declination 
    ra_tmp = ra_tmp / weightSum
    dec_tmp = dec_tmp / weightSum
    call sla_dtp2s(ra_tmp, dec_tmp, raz, decz, mergedsource%rightascension, mergedsource%declination)



    mergedsource%significance = significanceTemp / real(n)
    mergedsource%exposure = exposureTemp / real(n)
     

    !*****************************************************
    ! Go through and fill up the magnitude matrix
    !  
    ! loop over all the filters
    !******************************************************
    do j = 1 , nFilt


        nmag = 0
        nmag1 = 0
        significanceTemp = 0.0
        minorTemp = 0.0
        majorTemp = 0.0
        paTemp = 0.0

        !*********************************
        ! loop over all identical sources
        !*********************************
        do k = 1 , n

            if ( trim(sourceFilter(k)) .eq. trim(filterid(j)) ) then 
                nmag = nmag + 1      
                magTemp(nmag) = source(k)%magnitude
                magErrTemp(nmag) = source(k)%magnitudeErr
                corrRateTemp(nmag) = source(k)%corrCountRate
                corrRateErrTemp(nmag) = source(k)%corrCountRateErr
                significanceTemp = significanceTemp + source(k)%significance
                if (.not. isNullValue(source(k)%semia)) then
                    nmag1 = nmag1 + 1
                    majorTemp = majorTemp + source(k)%semiA
                    minorTemp = minorTemp + source(k)%semiB
                    paTemp = paTemp + source(k)%theta
                end if
                qflagTable(i, j) = source(k)%qualityFlag 
                cflagTable(i, j) = source(k)%confusionFlag
                eflagTable(i, j) = source(k)%extensionFlag
            end if
        end do

        !**********************************
        ! Compute the average values
        !**********************************
        if (nmag .lt. 1) then
!             write(*, *) "nmag=", nmag
             cycle
        end if
        

        call ComputeAverageMagnitude(magTemp, magErrTemp, nmag, mean, standardDeviation)        
        mag_table(i, j) = mean  
        magerr_table(i, j) = standardDeviation


        ! Compute the average corr-rate 
        call ComputeAverage(corrRateTemp, corrRateErrTemp, nmag, mean, standardDeviation)        
        corrRate_table(i, j) = mean  

        ! Compute errors from error propogation
        standardDeviation = sqrt(sum(corrRateErrTemp(1 : nMag) ** 2)) / real(nmag)    


        corrRateErr_table(i, j) = standardDeviation
!        omFitData(i)%counts(j) = mean
!        omFitData(i)%countserr(j) = standardDeviation 

        significanceTable(i, j) = significanceTemp / real(nmag)


!if (count .lt. 20) then

!write(*, *) i, " j=", j, " line 3861 n=", n, " nmag=", nmag, " significances=", significanceTable(i, :)
!count = count + 1
!else
!call fatal("Stopped at line 3864")
!end if

        if (nmag1 .gt. 0) then
            majorTable(i, j) = majorTemp / real(nmag1)          
            minorTable(i, j) = minorTemp / real(nmag1)        
            paTable(i, j)   = paTemp / real(nmag1)
        end if

        magexist(i, j) = .true.

    end do

end subroutine MergeSources
!*************************************************
!
!************************************************
Subroutine FindIdenticalSources(n, source, nfiles, nnew, index, sourceCount)
    integer, intent(in) :: n
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: nFiles
    integer(int32), pointer :: index(:, :)
    integer(int32), pointer :: sourceCount(:)
    logical, pointer :: matched(:) => null()
    integer :: i, j
    integer :: nmatched
 
    integer(int32) :: nnew   
    logical, pointer :: foundFile(:) => null()   
   
    do j = 1, nfiles
        do i = 1, n
            sourceCount(i) = 1
            index(i, j) = i
          
        end do
    end do

    if (nfiles .eq. 1) then ! Only one source-list file- nothing to do
        nnew = n
        do i = 1, n 
            source(i)%noSourcePixels = i
        end do                
        write(errstr, '(a, i6)') "Number of sources in combined source-list=", nnew
        call message(AppMsg, Verbose) 
        return
    end if


  
 
    allocate(foundfile(1 : nfiles))


    allocate(matched(1 : n))  ! Set to true when star used
    matched = .false.

    write(errstr,'(a, i6)') "Matching sources- no of input sources=", n
    call message(AppMsg, Verbose)
 

    nnew = 0 ! The number of sources in the merged source-list
    nmatched = 0
    
    matched = .false.
    sourceCount = 0
    index = 0
  
    !***********************************************************************
    ! Check for any sources with RAs near 0 and 360 degrees
    !**********************************************************************
    if( maxval(source%rightAscension) - minval(source%rightAscension) .gt. PI) then

        where(source%rightAscension .lt. PI) source%rightAscension = source%rightAscension + TWO_PI

    end if

    foundfile = .false.


    do i = 1, n
        if (matched(i)) cycle

        nmatched = 1
        foundfile= .false.
        foundfile(source(i)%imin) = .true.
          
        nnew = nnew + 1
        source(source(i)%number)%noSourcePixels = nnew  
             
        sourceCount(nnew) = 1    ! 
        index(nnew, 1) = i
         
        matched(i) = .true.

        ! i is the current index in the source-list
        ! n is the final index
        ! nnew is the current new source  


        call FindNeighbours(i, i + 1, n, nnew, nfiles, foundFile, sourceCount, &
                    index, source, matched)

          
    end do


 
    write(errstr, '(a, i6)') "Number of sources in combined source-list=", nnew
    call message(AppMsg, Verbose)
 

    deallocate(foundfile, matched)
    nullify(foundfile, matched)

    where(source%rightAscension .gt. TWO_PI) source%rightAscension = source%rightAscension - TWO_PI

end Subroutine FindIdenticalSources

!************************************************************
! Given a source, i, find all its neigbours,
! starting at index istart and finishing at n, the number
! of sources. Stars with matched(j) = true are skipped
!
!************************************************************
recursive subroutine FindNeighbours(i, istart, n, nnew, &
                     nfiles, foundFile, sourceCount, index, &
                     tempSource, matched)
    integer, intent(in) :: i, istart, n, nnew
    integer, intent(in) :: nfiles
    logical :: foundFile(:)
    integer(int32), pointer :: sourceCount(:), index(:, :)

    type(sourceParameters), pointer :: tempSource(:)
    logical, pointer :: matched(:)
  
    integer(int32) :: m
    real(single) :: minRadius
  
    if (sourceCount(nnew) .eq. nfiles) return

    do   ! Find all matches to source i

        if(.not. SourceMatch1(istart, n,  &
            tempSource(i), &
            tempsource,    &
            foundFile,     &
            tolerance, minRadius, matched, m)) return

                            
        sourceCount(nnew) = sourceCount(nnew) + 1   ! Number of sources comprising combined source nnew

        tempsource(tempSource(m)%number)%noSourcePixels = nnew
 
        index(nnew, sourcecount(nnew)) = m
       
        if (sourceCount(nnew) .eq. nfiles) return   ! Found all matches
   
        !******************************************
        ! Find any neighbours of this neighbour
        !******************************************        
        call FindNeighbours(m, 1, n, nnew, nfiles, foundFile,&
                            sourceCount, index, tempSource, matched)

        if (sourcecount(nnew) .eq. nfiles) return   ! Found all matches
   
    end do

end subroutine FindNeighbours

!*************************************************
!
!
!************************************************
Subroutine AlignCoordinateAxes(n, source, nfiles, nrows)
    integer, intent(in) :: n
    type(sourceParameters), pointer :: source(:)
  
    integer, intent(in) :: nFiles
    integer(int32) :: nrows(:)
    integer(int32), pointer, dimension(:) :: nStart => null(), nFinish => null()
    logical, pointer :: matched(:) => null()
    integer(int32), pointer :: raSortIndex(:) => null()
    integer(int32) :: i, j, k, l, m, i2(1)
    integer(int32) :: maxFile, nmatched, maxnmatched, minnmatched
    real(double) :: dx, dy, xoffset, yoffset, maxdx, maxdy
    real(double) :: sum1, sum2, sum3, sum4, error
    real(single) :: minX, maxX, minY, maxY   
    
    integer(int8), parameter :: noOffsets = 33
  
    real(single) :: offset(noOffsets) = (/0.0, -0.25, 0.25, -0.5, 0.5, -0.75, 0.75, -1.0, 1.0,       &  
                                        -1.25, 1.25, -1.5, 1.5, -1.75, 1.75, -2.0, 2.0, -2.25, 2.25, &
                                        -2.5, 2.5, -2.75, 2.75, -3.0, 3.0, -3.25, 3.25, -3.5, 3.5,   &
                                        -3.75, 3.75, -4.0, 4.0/)
                                 
    type(sourceParameters), pointer :: tempSource(:) => null()
    type(sourceParameters) :: tempSource1
    real(double) :: meanRadius, meanRadiusMin, weight, sumWeight 
    real(single) :: minRadius, nsigma1
    logical, pointer :: foundFile(:) => null() 
  
  
    if (nfiles .lt. 2) return
    allocate(foundfile(1 : nfiles))

    allocate(tempSource(1 : n)) ! For storing the combined source-list

    tempsource = source

    allocate(matched(1 : n))  ! Set to true when star used once
    matched = .false.

  

    offset = offset * degreesToRadians / 3600.0  ! Convert offsets to radians
    dx = 20.0 * degreesToRadians / 3600.0
    write(errstr,'(a)') "Attempting to align source coordinate axes in source-list files"
    call message(AppMsg, Verbose)
  
    nStart => allocateInt32(1, nFiles, "nstart")
    nFinish => allocateInt32(1, nFiles, "nfinish")
    
   
 

    ! Find the start and finish index of each source in the list for each file
   
    nStart(1) = 1
    nFinish(1) = nRows(1)

    
    do i = 2, nFiles
    
        nStart(i) = nFinish(i - 1) + 1
        nFinish(i) = nStart(i) + nRows(i) - 1
      
    end do

    k = 0     

    do i = 1, nfiles
    
        do j = 1, nrows(i)
            k = k + 1
            tempsource(k)%imin = i  ! Remember which file this source came from
            tempsource(k)%imax = j  ! Remember the row number in the source list file 
        end do
       
    end do
    

    !***********************************************************************
    ! Check for any sources with RAs near 0 and 360 degrees
    !**********************************************************************
    if( maxval(tempSource%rightAscension) - minval(tempSource%rightAscension) .gt. PI) then

        where(tempSource%rightAscension .lt. PI) tempSource%rightAscension = tempSource%rightAscension + TWO_PI

    end if

 
    ! The source-list file with the maximum number of stars
    i2 = maxloc(nrows(1 : nfiles))
    maxFile = i2(1)
  

    write(errstr, '(a, i3, a, i5, a)') "File ", maxFile, " has the maximum number of sources (", nrows(maxfile), ")"
    call message(AppMsg, Verbose)

    nsigma1 = nsigma
    nsigma = 2.0   
   
    minx = minval( tempSource( nstart(maxfile) : nfinish(maxfile))%rightAscension ) - dx
    maxx = maxval( tempSource( nstart(maxfile) : nfinish(maxfile))%rightAscension ) + dx
    miny = minval( tempSource( nstart(maxfile) : nfinish(maxfile))%declination ) - dx
    maxy = maxval( tempSource( nstart(maxfile) : nfinish(maxfile))%declination ) + dx


    !*****************************************************
    ! Sort the data for maxfile into RA increasing order
    !****************************************************
    raSortIndex => allocateInt32(1, nrows(maxfile), "raSortIndex")
    call IndexSort( source(nstart(maxfile) : nfinish(maxfile))%rightAscension, RASortIndex, nRows(maxFile))

    do i = 1, nrows(maxfile)
        tempSource(i + nstart(maxfile) - 1) = source( raSortIndex(i) + nstart(maxfile) - 1)
       
    end do

    deallocate(raSortIndex)
    nullify(raSortIndex)

 
       
    
    do i = 1, nfiles

        if (i .eq. maxfile) cycle 


        minnmatched = max(5, nrows(i) / 4)
        if (nrows(i) .lt. minnmatched) cycle
        maxnmatched = 0

           
        xoffset = 0.0
        yoffset = 0.0
        meanRadiusMin = 10.0 ! The minimum weighted sum of source separations
         
      
      jLoop :  do j = 1, noOffsets ! Loop through all the x offsets

            dx = offset(j)
            do k = 1, noOffsets    ! Loop through all the y offsets
 
                dy = offset(k)
                nmatched = 0
        
                meanRadius = 0.0_double
                sumWeight = 0.0_double

                matched = .false.
               
                sum1 = 0.0_double
                sum2 = 0.0_double
                sum3 = 0.0_double
                sum4 = 0.0_double
                do l = nstart(i), nfinish(i) ! Loop through the sources of file l and match them with the sources in file maxfile
                    if (matched(l)) cycle
                    
                    foundFile = .false.
                    tempSource1 = tempSource(l)
                    tempSource1%rightAscension = tempSource1%rightAscension + dx ! Add x offset to x coordinate
                    if (tempSource1%rightAscension .lt. minx) cycle 
                    if (tempSource1%rightAscension .gt. maxx) cycle 
               
                    tempSource1%declination = tempSource1%declination + dy       ! Add y offset to y coordinate
                    if (tempSource1%declination .lt. miny) cycle 
                    if (tempSource1%declination .gt. maxy) cycle 


                    !*******************************************************************
                    ! Try to find a match to tempsource1
                    !******************************************************************
                    if(SourceMatch1(nstart(maxfile), nfinish(maxfile), &
                           tempsource1,                                &
                           tempsource, foundfile, 2.0,                 &               
                           minRadius, matched, m)) then
                        error = AddErrors(dble(tempSource1%poserr), dble(tempSource(m)%poserr))
                

                        weight =  1.0_double / max( 0.01_double, error * error)
                        meanRadius = meanRadius + dble(minRadius) * weight
                        sumWeight = sumWeight + weight                                                  
                        nmatched = nmatched + 1 ! Number of source matched for the current offsets                 
                        
                        sum1 = sum1 + weight * tempSource(l)%rightAscension
                        sum2 = sum2 + weight * tempSource(l)%declination
                        sum3 = sum3 + weight * tempSource(m)%rightAscension
                        sum4 = sum4 + weight * tempSource(m)%declination


                    end if

                end do

                if (sumWeight .gt. 0.0_double) meanRadius = meanRadius / sumWeight 

                if (nmatched .ge. maxnmatched .and. meanRadius .lt. meanRadiusMin) then   ! Save the best overlapping offsets
                    maxnmatched = nmatched        
                    maxdx = dx
                    maxdy = dy
                    meanRadiusMin = meanRadius  
                    xoffset = (sum3 - sum1) / sumWeight
                    yoffset = (sum4 - sum2) / sumWeight                   
                                   
                end if
            end do


        end do  jLoop      
 
  
    

        if (maxnmatched .gt. minnmatched .and. meanRadiusMin .lt. 2.0) then

!            if(abs(xoffset - dx) .gt. 1.0) call fatal("wrong dx")
!            if(abs(yoffset - dy) .gt. 1.0) call fatal("wrong dy")

            write(errstr, '(a, i3, a, i3)') "Shifting source positions in file ", i, " to match those in file ", maxfile
            call message(AppMsg, Verbose)

         
            write(errstr, '(a, f5.2, a, f5.2, a, i4, a, i4)') "Approximate offsets - RA offset=", &
                       (maxdx * radiansToDegrees * 3600.0),                                   &
                       " arcsecs, Dec offset=", (maxdy * radiansToDegrees * 3600.0), &
                        " arcsecs, no matched=", maxnmatched, " no in file=", nrows(i)
            call message(AppMsg, Verbose)


            write(errstr, '(a, f5.2, a, f5.2, a)') "Least-squares offsets - RA offset=", &
                       (xoffset * radiansToDegrees * 3600.0),                                   &
                       " arcsecs, Dec offset=", (yoffset * radiansToDegrees * 3600.0), &
                        " arcsecs"
            call message(AppMsg, Verbose)

 
            !*************************************************************
            ! Correct the coordinates using the offsets
            !*************************************************************
            source(nstart(i) : nfinish(i))%rightascension = source(nstart(i) : nfinish(i))%rightascension + xoffset
            source(nstart(i) : nfinish(i))%declination = source(nstart(i) : nfinish(i))%declination + yoffset
        end if  
    end do

    nsigma = nsigma1

 
  
    deallocate(foundfile, nStart, nFinish, matched)
    nullify(foundfile, nStart, nFinish, matched)
   

end Subroutine AlignCoordinateAxes

!***************************************************************************
!
!
!**************************************************************************
subroutine AddMeanValTableColumn(tab, data, nData, nCols, colIndex, storeIndex, &
      heading, units, comment, nullValue)

    type(tableT), intent(inout) :: tab
    real(single), pointer :: data(:, :)
    integer, intent(in) :: nData, nCols
    integer, intent(in) :: colIndex(:), storeIndex(:)
    character(*), intent(in) :: heading, units, comment
    real(single), intent(in) :: nullValue
    integer :: i, j, k, l
    real(double) :: sum
    integer, dimension(1) :: s

    s(1) = 1

    do i = 1, nData
        k = 0
        sum = 0.0_double
        do j = 1, nCols
            l = colIndex(j)
            if (data(i, l) .ne. nullValue) then
                sum = sum + data(i, l)
                k = k + 1
            end if
        end do
        if (k .gt. 0) then
            if (k .eq. 1) then
                data(i, 1) = sum
            else
                data(i, 1) = sum / dble(k)
            end if
        end if
    end do

    call AddReal32IndexColumn(tab, heading, units, comment, nData, data(:, 1), storeIndex, nullValue)


end subroutine AddMeanValTableColumn



end module omsrclistcomb_mod









