
!*********************************************************************************************************** 
module sourcetable_mod

    use dal
    use cal
    use param  
    use sasdal
    
    use OmAllocation_mod         ! Functions for allocating memory to arrays.
    use OmSourceparam_mod             ! Source definitions, etc
    use omconstants_mod
    use OmPhotometryDefs_mod
    use omutility_mod   
    use errorHandling

    implicit none 
 
  private
  public CreateSourceListFile
contains 



!*********************************************
! Subroutine CreateSourceList
!*********************************************
subroutine CreateSourceListFile(outset,     &
                            nSources,       &
                            sourceDataIn,   &
                            exposure)
    character(len=*), intent(in) ::  outset

    integer(int32), intent(in) :: nSources 
    type(sourceParameters), pointer :: sourceDataIn(:)

 
    real(single), intent(in) :: exposure
   
      

    type(DataSetT) :: listDataSet 
    type(TableT) :: listTable
    type(ColumnT)        :: pcolumn                 ! pointer to a column

    integer (int32), pointer :: srcnum(:)  ! the source number
   
   
    integer(int32) :: i
   

    type(sourceParameters), pointer :: sourceData(:) => null()
write(*, *) exposure
    if (associated(sourceDataIn)) then
        sourceData => sourceDataIn
    else
       allocate(sourceData(1 : 1))
    end if

    ! Write out the source list to the oswDataSet
    write(errstr,'(a)') "Creating source list"
    call message(AppMsg, Verbose)

    !**********************************
    ! Create the output source list
    !**********************************
    listDataSet = dataSet(outset, create)

    listTable = addTable(listDataSet, "SRCLIST", nSources)

    
   
    ! this is an osw source list

    call setAttribute(listDataSet, "CONTENT", "OM OSW SOURCE LIST", &
                  "File contents")
    call setAttribute(listDataSet, "HDUCLASS", "OGIP", &
                  "Format conforms to OGIP/GSFC conventions")
    call setAttribute(listDataSet, "HDUCLAS1", "SRCLIST", &
                  "File contains a source list")

 
  

    !********************************
    ! now add the columns one by one
    ! starting with the source number
    !*********************************
    pcolumn = addColumn(listTable, "SRCNUM", Integer32, " ")
    srcnum => int32data(pcolumn)

    forall (i = 0 : nSources - 1) srcnum(i) = i + 1
   
    call release(pcolumn)


   
    ! Add the source positions
    call AddReal32Column(listTable, sourceData%xPos,  nSources, &
                         "XPOS", "pixels", "Centroided x-pixel position", nullValue)

    call AddReal32Column(listTable, sourceData%yPos,  nSources, &
                         "YPOS", "pixels", "Centroided y-pixel position", nullValue)

  
    call AddReal32Column(listTable, sourceData%poserr,  nSources, &
                         "POSERR", "pixels", "Centroided pixel position error", nullValue)
    


    call AddReal32Column(listTable,  sourceData%semiA, nSources, &
                 "FWHM_MAJ", "arcsecs", "FWHM of semi-major axis", nullValue)
 


    call AddReal32Column(listTable,  sourceData%semiB, nSources, &
                 "FWHM_MIN", "arcsecs", "FWHM of semi-minor axis", nullValue)

 
    call AddReal32Column(listTable,  sourceData%theta, nSources, &
                 "PA", "degrees", "PA of semi-major axis (measured anti-clockwise from x-axis)", nullValue)

    
    !************************
    ! release the memory
    !************************
    call release(listTable)
    call release(listDataSet)
    if (associated(sourceData) .and. .not. associated(sourceDataIn)) deallocate(sourceData) 

end subroutine CreateSourceListFile



end module sourcetable_mod








